#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import time
import uuid
import os
from unittest import mock
from parlai.mturk.core.socket_manager import Packet, SocketManager
from parlai.mturk.core.shared_utils import AssignState
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.core.params import ParlaiParser

import parlai.mturk.core.mturk_manager as MTurkManagerFile
import parlai.mturk.core.data_model as data_model
import parlai.mturk.core.shared_utils as shared_utils
import threading
from websocket_server import WebsocketServer
import json
import parlai.utils.testing as testing_utils

parent_dir = os.path.dirname(os.path.abspath(__file__))
MTurkManagerFile.parent_dir = os.path.dirname(os.path.abspath(__file__))

# Lets ignore the logging part
MTurkManagerFile.shared_utils.print_and_log = mock.MagicMock()

TEST_WORKER_ID_1 = 'TEST_WORKER_ID_1'
TEST_WORKER_ID_2 = 'TEST_WORKER_ID_2'
TEST_ASSIGNMENT_ID_1 = 'TEST_ASSIGNMENT_ID_1'
TEST_ASSIGNMENT_ID_2 = 'TEST_ASSIGNMENT_ID_2'
TEST_ASSIGNMENT_ID_3 = 'TEST_ASSIGNMENT_ID_3'
TEST_HIT_ID_1 = 'TEST_HIT_ID_1'
TEST_HIT_ID_2 = 'TEST_HIT_ID_2'
TEST_CONV_ID_1 = 'TEST_CONV_ID_1'
FAKE_ID = 'BOGUS'

MESSAGE_ID_1 = 'MESSAGE_ID_1'
MESSAGE_ID_2 = 'MESSAGE_ID_2'
MESSAGE_ID_3 = 'MESSAGE_ID_3'
MESSAGE_ID_4 = 'MESSAGE_ID_4'
COMMAND_ID_1 = 'COMMAND_ID_1'

MESSAGE_TYPE = data_model.MESSAGE_TYPE_MESSAGE
COMMAND_TYPE = data_model.MESSAGE_TYPE_COMMAND

MESSAGE_1 = {'message_id': MESSAGE_ID_1, 'type': MESSAGE_TYPE}
MESSAGE_2 = {'message_id': MESSAGE_ID_2, 'type': MESSAGE_TYPE}
COMMAND_1 = {'message_id': COMMAND_ID_1, 'type': COMMAND_TYPE}

AGENT_ID = 'AGENT_ID'

ACT_1 = {'text': 'THIS IS A MESSAGE', 'id': AGENT_ID}
ACT_2 = {'text': 'THIS IS A MESSAGE AGAIN', 'id': AGENT_ID}

active_statuses = [
    AssignState.STATUS_NONE,
    AssignState.STATUS_ONBOARDING,
    AssignState.STATUS_WAITING,
    AssignState.STATUS_IN_TASK,
]
complete_statuses = [
    AssignState.STATUS_DONE,
    AssignState.STATUS_DISCONNECT,
    AssignState.STATUS_PARTNER_DISCONNECT,
    AssignState.STATUS_PARTNER_DISCONNECT_EARLY,
    AssignState.STATUS_EXPIRED,
    AssignState.STATUS_RETURNED,
]
statuses = active_statuses + complete_statuses

TASK_GROUP_ID_1 = 'TASK_GROUP_ID_1'

SocketManager.DEF_MISSED_PONGS = 1
SocketManager.HEARTBEAT_RATE = 0.4
SocketManager.DEF_DEAD_TIME = 0.4
SocketManager.ACK_TIME = {Packet.TYPE_ALIVE: 0.4, Packet.TYPE_MESSAGE: 0.2}

shared_utils.THREAD_SHORT_SLEEP = 0.05
shared_utils.THREAD_MEDIUM_SLEEP = 0.15

MTurkManagerFile.WORLD_START_TIMEOUT = 2


TOPIC_ARN = 'topic_arn'
QUALIFICATION_ID = 'qualification_id'
HIT_TYPE_ID = 'hit_type_id'
MTURK_PAGE_URL = 'mturk_page_url'
FAKE_HIT_ID = 'fake_hit_id'


def assert_equal_by(val_func, val, max_time):
    start_time = time.time()
    while val_func() != val:
        assert (
            time.time() - start_time < max_time
        ), "Value was not attained in specified time"
        time.sleep(0.1)


class MockSocket:
    def __init__(self):
        self.last_messages = {}
        self.connected = False
        self.disconnected = False
        self.closed = False
        self.ws = None
        self.should_heartbeat = True
        self.fake_workers = []
        self.port = None
        self.launch_socket()
        self.handlers = {}
        while self.ws is None:
            time.sleep(0.05)
        time.sleep(1)

    def send(self, packet):
        self.ws.send_message_to_all(packet)

    def close(self):
        if not self.closed:
            self.ws.server_close()
            self.ws.shutdown()
            self.closed = True

    def do_nothing(self, *args):
        pass

    def launch_socket(self):
        def on_message(client, server, message):
            if self.closed:
                raise Exception('Socket is already closed...')
            if message == '':
                return
            packet_dict = json.loads(message)
            if packet_dict['content']['id'] == 'WORLD_ALIVE':
                self.ws.send_message(client, json.dumps({'type': 'conn_success'}))
                self.connected = True
            elif packet_dict['content']['type'] == 'heartbeat':
                pong = packet_dict['content'].copy()
                pong['type'] = 'pong'
                self.ws.send_message(
                    client,
                    json.dumps(
                        {'type': data_model.SOCKET_ROUTE_PACKET_STRING, 'content': pong}
                    ),
                )
            if 'receiver_id' in packet_dict['content']:
                receiver_id = packet_dict['content']['receiver_id']
                assignment_id = packet_dict['content']['assignment_id']
                use_func = self.handlers.get(
                    receiver_id + assignment_id, self.do_nothing
                )
                use_func(packet_dict['content'])

        def on_connect(client, server):
            pass

        def on_disconnect(client, server):
            self.disconnected = True

        def run_socket(*args):
            port = 3030
            while self.port is None:
                try:
                    self.ws = WebsocketServer(port, host='127.0.0.1')
                    self.port = port
                except OSError:
                    port += 1
            self.ws.set_fn_client_left(on_disconnect)
            self.ws.set_fn_new_client(on_connect)
            self.ws.set_fn_message_received(on_message)
            self.ws.run_forever()

        self.listen_thread = threading.Thread(
            target=run_socket, name='Fake-Socket-Thread'
        )
        self.listen_thread.daemon = True
        self.listen_thread.start()


class MockAgent(object):
    """
    Class that pretends to be an MTurk agent interacting through the webpage by
    simulating the same commands that are sent from the core.html file.

    Exposes methods to use for testing and checking status
    """

    def __init__(self, hit_id, assignment_id, worker_id, task_group_id):
        self.conversation_id = None
        self.id = None
        self.assignment_id = assignment_id
        self.hit_id = hit_id
        self.worker_id = worker_id
        self.some_agent_disconnected = False
        self.disconnected = False
        self.task_group_id = task_group_id
        self.ws = None
        self.always_beat = True
        self.send_acks = True
        self.ready = False
        self.wants_to_send = False
        self.acked_packet = []
        self.incoming_hb = []
        self.message_packet = []

    def send_packet(self, packet):
        def callback(*args):
            pass

        event_name = data_model.SOCKET_ROUTE_PACKET_STRING
        self.ws.send(json.dumps({'type': event_name, 'content': packet.as_dict()}))

    def register_to_socket(self, ws):
        handler = self.make_packet_handler()
        self.ws = ws
        self.ws.handlers[self.worker_id + self.assignment_id] = handler

    def on_msg(self, packet):
        self.message_packet.append(packet)
        if packet.data['text'] == data_model.COMMAND_CHANGE_CONVERSATION:
            self.ready = False
            self.conversation_id = packet.data['conversation_id']
            self.id = packet.data['agent_id']
            self.send_alive()

    def make_packet_handler(self):
        """
        A packet handler that properly sends heartbeats.
        """

        def on_ack(*args):
            self.acked_packet.append(args[0])

        def on_hb(*args):
            self.incoming_hb.append(args[0])

        def handler_mock(pkt):
            if pkt['type'] == Packet.TYPE_ACK:
                self.ready = True
                packet = Packet.from_dict(pkt)
                on_ack(packet)
            elif pkt['type'] == Packet.TYPE_HEARTBEAT:
                packet = Packet.from_dict(pkt)
                on_hb(packet)
                if self.always_beat:
                    self.send_heartbeat()
            elif pkt['type'] == Packet.TYPE_MESSAGE:
                packet = Packet.from_dict(pkt)
                if self.send_acks:
                    self.send_packet(packet.get_ack())
                self.on_msg(packet)
            elif pkt['type'] == Packet.TYPE_ALIVE:
                raise Exception('Invalid alive packet {}'.format(pkt))
            else:
                raise Exception(
                    'Invalid Packet type {} received in {}'.format(pkt['type'], pkt)
                )

        return handler_mock

    def build_and_send_packet(self, packet_type, data):
        msg = {
            'id': str(uuid.uuid4()),
            'type': packet_type,
            'sender_id': self.worker_id,
            'assignment_id': self.assignment_id,
            'conversation_id': self.conversation_id,
            'receiver_id': '[World_' + self.task_group_id + ']',
            'data': data,
        }

        event_name = data_model.SOCKET_ROUTE_PACKET_STRING
        if packet_type == Packet.TYPE_ALIVE:
            event_name = data_model.SOCKET_AGENT_ALIVE_STRING
        self.ws.send(json.dumps({'type': event_name, 'content': msg}))
        return msg['id']

    def send_message(self, text):
        data = {
            'text': text,
            'id': self.id,
            'message_id': str(uuid.uuid4()),
            'episode_done': False,
        }

        self.wants_to_send = False
        return self.build_and_send_packet(Packet.TYPE_MESSAGE, data)

    def send_alive(self):
        data = {
            'hit_id': self.hit_id,
            'assignment_id': self.assignment_id,
            'worker_id': self.worker_id,
            'conversation_id': self.conversation_id,
        }
        return self.build_and_send_packet(Packet.TYPE_ALIVE, data)

    def send_heartbeat(self):
        """
        Sends a heartbeat to the world.
        """
        hb = {
            'id': str(uuid.uuid4()),
            'receiver_id': '[World_' + self.task_group_id + ']',
            'assignment_id': self.assignment_id,
            'sender_id': self.worker_id,
            'conversation_id': self.conversation_id,
            'type': Packet.TYPE_HEARTBEAT,
            'data': None,
        }
        self.ws.send(
            json.dumps({'type': data_model.SOCKET_ROUTE_PACKET_STRING, 'content': hb})
        )

    def wait_for_alive(self):
        last_time = time.time()
        while not self.ready:
            self.send_alive()
            time.sleep(0.5)
            assert (
                time.time() - last_time < 10
            ), 'Timed out wating for server to acknowledge {} alive'.format(
                self.worker_id
            )


class TestMTurkManagerWorkflows(unittest.TestCase):
    """
    Various test cases to replicate a whole mturk workflow.
    """

    def setUp(self):
        patcher = mock.patch('builtins.input', return_value='y')
        self.addCleanup(patcher.stop)
        patcher.start()
        # Mock functions that hit external APIs and such
        self.server_utils = MTurkManagerFile.server_utils
        self.mturk_utils = MTurkManagerFile.mturk_utils
        self.server_utils.setup_server = mock.MagicMock(
            return_value='https://127.0.0.1'
        )
        self.server_utils.setup_legacy_server = mock.MagicMock(
            return_value='https://127.0.0.1'
        )
        self.server_utils.delete_server = mock.MagicMock()
        self.mturk_utils.setup_aws_credentials = mock.MagicMock()
        self.mturk_utils.calculate_mturk_cost = mock.MagicMock(return_value=1)
        self.mturk_utils.check_mturk_balance = mock.MagicMock(return_value=True)
        self.mturk_utils.create_hit_config = mock.MagicMock()
        self.mturk_utils.setup_sns_topic = mock.MagicMock(return_value=TOPIC_ARN)
        self.mturk_utils.delete_sns_topic = mock.MagicMock()
        self.mturk_utils.delete_qualification = mock.MagicMock()
        self.mturk_utils.find_or_create_qualification = mock.MagicMock(
            return_value=QUALIFICATION_ID
        )
        self.mturk_utils.find_qualification = mock.MagicMock(
            return_value=QUALIFICATION_ID
        )
        self.mturk_utils.give_worker_qualification = mock.MagicMock()
        self.mturk_utils.remove_worker_qualification = mock.MagicMock()
        self.mturk_utils.create_hit_type = mock.MagicMock(return_value=HIT_TYPE_ID)
        self.mturk_utils.subscribe_to_hits = mock.MagicMock()
        self.mturk_utils.create_hit_with_hit_type = mock.MagicMock(
            return_value=(MTURK_PAGE_URL, FAKE_HIT_ID, 'MTURK_HIT_DATA')
        )
        self.mturk_utils.get_mturk_client = mock.MagicMock(
            return_value=mock.MagicMock()
        )

        self.onboarding_agents = {}
        self.worlds_agents = {}

        # Set up an MTurk Manager and get it ready for accepting workers
        self.fake_socket = MockSocket()
        time.sleep(0.1)
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 1
        self.opt['hit_title'] = 'test_hit_title'
        self.opt['hit_description'] = 'test_hit_description'
        self.opt['task_description'] = 'test_task_description'
        self.opt['hit_keywords'] = 'test_hit_keywords'
        self.opt['reward'] = 0.1
        self.opt['is_debug'] = True
        self.opt['log_level'] = 0
        self.opt['num_conversations'] = 1
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        self.mturk_manager.port = self.fake_socket.port
        self.mturk_manager.setup_server()
        self.mturk_manager.start_new_run()
        self.mturk_manager.ready_to_accept_workers()
        self.mturk_manager.set_onboard_function(self.onboard_agent)
        self.mturk_manager.create_hits()

        def assign_worker_roles(workers):
            workers[0].id = 'mturk_agent_1'
            workers[1].id = 'mturk_agent_2'

        def run_task_wait():
            self.mturk_manager.start_task(
                lambda w: True, assign_worker_roles, self.run_conversation
            )

        self.task_thread = threading.Thread(target=run_task_wait)
        self.task_thread.start()

        self.agent_1 = MockAgent(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_1, TEST_WORKER_ID_1, TASK_GROUP_ID_1
        )
        self.agent_1_2 = MockAgent(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_3, TEST_WORKER_ID_1, TASK_GROUP_ID_1
        )
        self.agent_2 = MockAgent(
            TEST_HIT_ID_2, TEST_ASSIGNMENT_ID_2, TEST_WORKER_ID_2, TASK_GROUP_ID_1
        )

    def tearDown(self):
        self.agent_1.always_beat = False
        self.agent_2.always_beat = False
        for key in self.worlds_agents.keys():
            self.worlds_agents[key] = True
        self.mturk_manager.shutdown()
        self.fake_socket.close()
        self.task_thread.join()

    def onboard_agent(self, worker):
        self.onboarding_agents[worker.worker_id] = False
        while (worker.worker_id in self.onboarding_agents) and (
            self.onboarding_agents[worker.worker_id] is False
        ):
            time.sleep(0.05)
        return

    def run_conversation(self, mturk_manager, opt, workers):
        for worker in workers:
            self.worlds_agents[worker.worker_id] = False
        for worker in workers:
            while self.worlds_agents[worker.worker_id] is False:
                time.sleep(0.05)
        for worker in workers:
            worker.shutdown(timeout=-1)

    def alive_agent(self, agent):
        agent.register_to_socket(self.fake_socket)
        agent.wait_for_alive()
        agent.send_heartbeat()

    @testing_utils.retry()
    def test_successful_convo(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 2)

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        # Assert conversation is complete for manager and agents
        assert_equal_by(lambda: manager.completed_conversations, 1, 2)
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_2.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 2, 2
        )

    @testing_utils.retry()
    def test_disconnect_end(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 2)

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Disconnect agent
        agent_2.always_beat = False
        assert_equal_by(
            agent_1_object.get_status, AssignState.STATUS_PARTNER_DISCONNECT, 3
        )
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DISCONNECT, 3)
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        agent_2.always_beat = True
        agent_2.send_alive()

        # Assert workers get the correct command
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_INACTIVE_DONE
                ]
            ),
            1,
            2,
        )
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_2.message_packet
                    if p.data['text'] == data_model.COMMAND_INACTIVE_HIT
                ]
            ),
            1,
            2,
        )

        # assert conversation not marked as complete
        self.assertEqual(manager.completed_conversations, 0)

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 2, 2
        )

    @testing_utils.retry()
    def test_expire_onboarding(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 10)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)

        manager._expire_onboarding_pool()

        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_EXPIRE_HIT
                ]
            ),
            1,
            10,
        )

        self.onboarding_agents[agent_1.worker_id] = True

        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_EXPIRED)

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]),
            1,
            10,
        )

    @testing_utils.retry()
    def test_reconnect_complete(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 2)

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Simulate reconnect to task
        stored_conv_id = agent_2.conversation_id
        stored_agent_id = agent_2.id
        agent_2.conversation_id = None
        agent_2.id = None
        agent_2.send_alive()

        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_2.message_packet
                    if p.data['text'] == data_model.COMMAND_RESTORE_STATE
                ]
            ),
            1,
            4,
        )
        self.assertEqual(agent_2.id, stored_agent_id)
        self.assertEqual(agent_2.conversation_id, stored_conv_id)

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        # Assert conversation is complete for manager and agents
        assert_equal_by(lambda: manager.completed_conversations, 1, 2)
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_2.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 2, 2
        )

    @testing_utils.retry()
    def test_attempt_break_unique(self):
        manager = self.mturk_manager
        unique_worker_qual = 'is_unique_qual'
        manager.is_unique = True
        manager.opt['unique_qual_name'] = unique_worker_qual
        manager.unique_qual_name = unique_worker_qual

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 2)

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        # Assert conversation is complete for manager and agents
        assert_equal_by(lambda: manager.completed_conversations, 1, 2)
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_2.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 2, 2
        )

        # ensure qualification was 'granted'
        self.mturk_utils.find_qualification.assert_called_with(
            unique_worker_qual, manager.is_sandbox
        )
        self.mturk_utils.give_worker_qualification.assert_any_call(
            agent_1.worker_id, QUALIFICATION_ID, None, manager.is_sandbox
        )
        self.mturk_utils.give_worker_qualification.assert_any_call(
            agent_2.worker_id, QUALIFICATION_ID, None, manager.is_sandbox
        )

        # Try to alive with the first agent a second time
        agent_1_2 = self.agent_1_2
        self.alive_agent(agent_1_2)
        assert_equal_by(lambda: agent_1_2.worker_id in self.onboarding_agents, True, 2)
        agent_1_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_1_2.assignment_id
        )

        # No worker should be created for a unique task
        self.assertIsNone(agent_1_2_object)

        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1_2.message_packet
                    if p.data['text'] == data_model.COMMAND_EXPIRE_HIT
                ]
            ),
            1,
            2,
        )

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 3, 2
        )

    @testing_utils.retry()
    def test_break_multi_convo(self):
        manager = self.mturk_manager
        manager.opt['allowed_conversations'] = 1

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 2)

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Attempt to start a new conversation with duplicate worker 1
        agent_1_2 = self.agent_1_2
        self.alive_agent(agent_1_2)
        assert_equal_by(lambda: agent_1_2.worker_id in self.onboarding_agents, True, 2)
        agent_1_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_1_2.assignment_id
        )

        # No worker should be created for a unique task
        self.assertIsNone(agent_1_2_object)

        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1_2.message_packet
                    if p.data['text'] == data_model.COMMAND_EXPIRE_HIT
                ]
            ),
            1,
            2,
        )

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 1, 2
        )

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        # Assert conversation is complete for manager and agents
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )
        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_2.message_packet
                    if p.data['text'] == data_model.COMMAND_SHOW_DONE_BUTTON
                ]
            ),
            1,
            2,
        )
        assert_equal_by(lambda: manager.completed_conversations, 1, 2)

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 3, 2
        )

    @testing_utils.retry()
    def test_no_onboard_expire_waiting(self):
        manager = self.mturk_manager
        manager.set_onboard_function(None)

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        manager._expire_agent_pool()

        assert_equal_by(
            lambda: len(
                [
                    p
                    for p in agent_1.message_packet
                    if p.data['text'] == data_model.COMMAND_EXPIRE_HIT
                ]
            ),
            1,
            2,
        )

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 1, 2
        )

    @testing_utils.retry()
    def test_return_to_waiting_on_world_start(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Make agent_1 no longer respond to change_conversation_requests
        def replace_on_msg(packet):
            agent_1.message_packet.append(packet)

        agent_1.on_msg = replace_on_msg

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 2)

        # Assert agents attempt to move to task, but then move back to waiting
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_WAITING, 3)
        agent_1.always_beat = False

        # Assert no world ever started
        self.assertNotIn(agent_2.worker_id, self.worlds_agents)

        # Expire everything
        manager.shutdown()

        # Assert sockets are closed
        assert_equal_by(
            lambda: len([x for x in manager.socket_manager.run.values() if not x]), 2, 2
        )


if __name__ == '__main__':
    unittest.main(buffer=True)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import os
import time
import uuid
from datetime import datetime

from parlai.mturk.core.mturk_data_handler import MTurkDataHandler
from parlai.mturk.core.shared_utils import AssignState

import parlai.mturk.core.mturk_data_handler as DataHandlerFile

data_dir = os.path.dirname(os.path.abspath(__file__))
DataHandlerFile.data_dir = data_dir


class TestDataHandler(unittest.TestCase):
    """
    Various unit tests for the SQLite database.
    """

    DB_NAME = 'test_db.db'

    def setUp(self):
        if os.path.exists(os.path.join(data_dir, self.DB_NAME)):
            os.remove(os.path.join(data_dir, self.DB_NAME))

    def tearDown(self):
        if os.path.exists(os.path.join(data_dir, self.DB_NAME)):
            os.remove(os.path.join(data_dir, self.DB_NAME))

    def create_hit(self):
        return {
            'HIT': {
                'HITId': str(uuid.uuid4()),
                'Expiration': datetime.today().replace(microsecond=0),
                'HITStatus': str(uuid.uuid4()),
                'NumberOfAssignmentsPending': 1,
                'NumberOfAssignmentsAvailable': 2,
                'NumberOfAssignmentsCompleted': 3,
            }
        }

    def assertHITEqual(self, mturk_hit, db_hit, run_id):
        self.assertEqual(mturk_hit['HIT']['HITId'], db_hit['hit_id'])
        self.assertEqual(
            mturk_hit['HIT']['Expiration'], datetime.fromtimestamp(db_hit['expiration'])
        )
        self.assertEqual(mturk_hit['HIT']['HITStatus'], db_hit['hit_status'])
        self.assertEqual(
            mturk_hit['HIT']['NumberOfAssignmentsPending'],
            db_hit['assignments_pending'],
        )
        self.assertEqual(
            mturk_hit['HIT']['NumberOfAssignmentsAvailable'],
            db_hit['assignments_available'],
        )
        self.assertEqual(
            mturk_hit['HIT']['NumberOfAssignmentsCompleted'],
            db_hit['assignments_complete'],
        )
        self.assertEqual(run_id, db_hit['run_id'])

    def test_init_db(self):
        db_logger = MTurkDataHandler('test1', file_name=self.DB_NAME)
        conn = db_logger._get_connection()
        c = conn.cursor()
        c.execute('SELECT COUNT(*) FROM runs;')
        self.assertEqual(c.fetchone()[0], 0)
        c.execute('SELECT COUNT(*) FROM hits;')
        self.assertEqual(c.fetchone()[0], 0)
        c.execute('SELECT COUNT(*) FROM assignments;')
        self.assertEqual(c.fetchone()[0], 0)
        c.execute('SELECT COUNT(*) FROM workers;')
        self.assertEqual(c.fetchone()[0], 0)
        c.execute('SELECT COUNT(*) FROM pairings;')
        self.assertEqual(c.fetchone()[0], 0)

    def test_create_get_run(self):
        run_id = 'Test_run_1'
        hits_created = 10
        db_logger = MTurkDataHandler('test2', file_name=self.DB_NAME)

        # Ensure a run logged can be retrieved
        db_logger.log_new_run(hits_created, 'testing', run_id)
        run_data = db_logger.get_run_data(run_id)
        self.assertEqual(run_data['run_id'], run_id)
        self.assertEqual(run_data['created'], 0)
        self.assertEqual(run_data['completed'], 0)
        self.assertEqual(run_data['maximum'], hits_created)
        self.assertEqual(run_data['failed'], 0)

        # Assert missed entries are None
        self.assertIsNone(db_logger.get_run_data('fake_id'))

    def test_create_update_hits(self):
        run_id = 'Test_run_2'
        hits_created = 10
        db_logger = MTurkDataHandler(file_name=self.DB_NAME)
        db_logger.log_new_run(hits_created, 'testing', run_id)
        HIT1 = self.create_hit()
        HIT2 = self.create_hit()
        HIT3 = self.create_hit()

        # Ensure logging without group id fails
        with self.assertRaises(AssertionError):
            db_logger.log_hit_status(HIT1)

        # Log created hits through one logger
        db_logger.log_hit_status(HIT1, run_id)
        db_logger.log_hit_status(HIT2, run_id)

        # Create new handler, this one with the group id created, ensure
        # the log works fine
        db_logger = MTurkDataHandler(run_id, file_name=self.DB_NAME)
        db_logger.log_hit_status(HIT3)

        # Ensure all of the expected hits are there
        run_data = db_logger.get_run_data(run_id)
        self.assertEqual(run_data['run_id'], run_id)
        self.assertEqual(run_data['created'], 3)
        self.assertEqual(run_data['completed'], 0)
        self.assertEqual(run_data['maximum'], hits_created)
        self.assertEqual(run_data['failed'], 0)

        # Ensure the hit details are correct
        for hit in [HIT1, HIT2, HIT3]:
            hit_db_data = db_logger.get_hit_data(hit['HIT']['HITId'])
            self.assertHITEqual(hit, hit_db_data, run_id)

        # Update the data on a HIT, ensure that the run data stays the same
        # but the HIT data updates
        test_status = 'TEST_STATUS'
        HIT2['HIT']['HITStatus'] = test_status
        db_logger.log_hit_status(HIT2)

        # Ensure all of the expected hits are there
        run_data = db_logger.get_run_data(run_id)
        self.assertEqual(run_data['run_id'], run_id)
        self.assertEqual(run_data['created'], 3)
        self.assertEqual(run_data['completed'], 0)
        self.assertEqual(run_data['maximum'], hits_created)
        self.assertEqual(run_data['failed'], 0)

        # Ensure the hit details are correct
        for hit in [HIT1, HIT2, HIT3]:
            hit_db_data = db_logger.get_hit_data(hit['HIT']['HITId'])
            self.assertHITEqual(hit, hit_db_data, run_id)

        # Ensure requesting a hit that doesn't exist returns none
        self.assertIsNone(db_logger.get_hit_data('fake_id'))

    def test_worker_workflows(self):
        run_id = 'Test_run_3'
        hits_created = 10
        db_logger = MTurkDataHandler(run_id, file_name=self.DB_NAME)
        db_logger.log_new_run(hits_created, 'testing', run_id)
        HIT1 = self.create_hit()
        HIT2 = self.create_hit()
        HIT3 = self.create_hit()
        db_logger.log_hit_status(HIT1)
        db_logger.log_hit_status(HIT2)
        db_logger.log_hit_status(HIT3)

        worker_id_1 = 'TEST_WORKER_ID_1'
        worker_id_2 = 'TEST_WORKER_ID_2'
        assignment_id_1 = 'TEST_ASSIGNMENT_ID_1'
        assignment_id_2 = 'TEST_ASSIGNMENT_ID_2'
        assignment_id_3 = 'TEST_ASSIGNMENT_ID_3'

        # Create two workers and assign the 3 assignments to them
        db_logger.log_worker_accept_assignment(
            worker_id_1, assignment_id_1, HIT1['HIT']['HITId']
        )
        db_logger.log_worker_accept_assignment(
            worker_id_2, assignment_id_2, HIT2['HIT']['HITId']
        )
        db_logger.log_worker_accept_assignment(
            worker_id_2, assignment_id_3, HIT3['HIT']['HITId']
        )

        # Ensure two workers have been created
        conn = db_logger._get_connection()
        c = conn.cursor()
        c.execute('SELECT COUNT(*) FROM workers;')
        self.assertEqual(c.fetchone()[0], 2)

        # Ensure non-existent worker is None
        self.assertIsNone(db_logger.get_worker_data('fake_id'))

        # Ensure the two workers have the correct expected values
        worker_1_data = db_logger.get_worker_data(worker_id_1)
        worker_2_data = db_logger.get_worker_data(worker_id_2)
        self.assertEqual(worker_1_data['worker_id'], worker_id_1)
        self.assertEqual(worker_1_data['accepted'], 1)
        self.assertEqual(worker_1_data['disconnected'], 0)
        self.assertEqual(worker_1_data['completed'], 0)
        self.assertEqual(worker_1_data['approved'], 0)
        self.assertEqual(worker_1_data['rejected'], 0)
        self.assertEqual(worker_2_data['worker_id'], worker_id_2)
        self.assertEqual(worker_2_data['accepted'], 2)
        self.assertEqual(worker_2_data['disconnected'], 0)
        self.assertEqual(worker_2_data['completed'], 0)
        self.assertEqual(worker_2_data['approved'], 0)
        self.assertEqual(worker_2_data['rejected'], 0)

        # Ensure all the assignments are marked as accepted
        c.execute('SELECT COUNT(*) FROM assignments WHERE status = ?;', ('Accepted',))
        self.assertEqual(c.fetchone()[0], 3)

        # Ensure non-existing assign is None
        self.assertIsNone(db_logger.get_assignment_data('fake_id'))

        # Check each of the assignments
        assignment_1_data = db_logger.get_assignment_data(assignment_id_1)
        assignment_2_data = db_logger.get_assignment_data(assignment_id_2)
        assignment_3_data = db_logger.get_assignment_data(assignment_id_3)
        self.assertEqual(assignment_1_data['assignment_id'], assignment_id_1)
        self.assertEqual(assignment_1_data['status'], 'Accepted')
        self.assertEqual(assignment_1_data['approve_time'], None)
        self.assertEqual(assignment_1_data['worker_id'], worker_id_1)
        self.assertEqual(assignment_1_data['hit_id'], HIT1['HIT']['HITId'])
        self.assertEqual(assignment_2_data['assignment_id'], assignment_id_2)
        self.assertEqual(assignment_2_data['status'], 'Accepted')
        self.assertEqual(assignment_2_data['approve_time'], None)
        self.assertEqual(assignment_2_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_2_data['hit_id'], HIT2['HIT']['HITId'])
        self.assertEqual(assignment_3_data['assignment_id'], assignment_id_3)
        self.assertEqual(assignment_3_data['status'], 'Accepted')
        self.assertEqual(assignment_3_data['approve_time'], None)
        self.assertEqual(assignment_3_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_3_data['hit_id'], HIT3['HIT']['HITId'])

        # Ensure three pairings have been created, one for each assignment
        c.execute('SELECT COUNT(*) FROM pairings')
        self.assertEqual(c.fetchone()[0], 3)

        # Ensure pairings are accurate
        self.assertIsNone(
            db_logger.get_worker_assignment_pairing(worker_id_1, assignment_id_3)
        )

        pair_1 = db_logger.get_worker_assignment_pairing(worker_id_1, assignment_id_1)
        pair_2 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_2)
        pair_3 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_3)
        for f in [
            'onboarding_start',
            'onboarding_end',
            'task_start',
            'task_end',
            'conversation_id',
        ]:
            for pair in [pair_1, pair_2, pair_3]:
                self.assertIsNone(pair[f])
        self.assertEqual(pair_1['status'], AssignState.STATUS_NONE)
        self.assertEqual(pair_2['status'], AssignState.STATUS_NONE)
        self.assertEqual(pair_3['status'], AssignState.STATUS_NONE)
        self.assertEqual(pair_1['worker_id'], worker_id_1)
        self.assertEqual(pair_2['worker_id'], worker_id_2)
        self.assertEqual(pair_3['worker_id'], worker_id_2)
        self.assertEqual(pair_1['assignment_id'], assignment_id_1)
        self.assertEqual(pair_2['assignment_id'], assignment_id_2)
        self.assertEqual(pair_3['assignment_id'], assignment_id_3)
        self.assertEqual(pair_1['run_id'], run_id)
        self.assertEqual(pair_2['run_id'], run_id)
        self.assertEqual(pair_3['run_id'], run_id)
        self.assertEqual(pair_1['bonus_amount'], 0)
        self.assertEqual(pair_2['bonus_amount'], 0)
        self.assertEqual(pair_3['bonus_amount'], 0)
        self.assertEqual(pair_1['bonus_text'], '')
        self.assertEqual(pair_2['bonus_text'], '')
        self.assertEqual(pair_3['bonus_text'], '')
        self.assertFalse(pair_1['bonus_paid'])
        self.assertFalse(pair_2['bonus_paid'])
        self.assertFalse(pair_3['bonus_paid'])

        # Ensure get_pairings_for_assignment works
        pair_4 = db_logger.get_pairings_for_assignment(assignment_id_2)[0]
        self.assertEqual(pair_2, pair_4)
        self.assertListEqual([], db_logger.get_pairings_for_assignment('fake_id'))

        # Ensure get_all_<thing>_for_worker works
        self.assertListEqual([], db_logger.get_all_assignments_for_worker('fake_id'))
        self.assertListEqual([], db_logger.get_all_pairings_for_worker('fake_id'))
        self.assertEqual(
            db_logger.get_all_assignments_for_worker(worker_id_1)[0], assignment_1_data
        )
        self.assertEqual(len(db_logger.get_all_assignments_for_worker(worker_id_2)), 2)
        self.assertEqual(db_logger.get_all_pairings_for_worker(worker_id_1)[0], pair_1)
        self.assertEqual(len(db_logger.get_all_pairings_for_worker(worker_id_2)), 2)

        # test task_restricted gets
        self.assertEqual(
            db_logger.get_all_task_assignments_for_worker(worker_id_1)[0],
            assignment_1_data,
        )
        self.assertEqual(
            len(db_logger.get_all_task_assignments_for_worker(worker_id_2)), 2
        )
        self.assertEqual(
            len(db_logger.get_all_task_assignments_for_worker(worker_id_1, 'fake_id')),
            0,
        )
        self.assertEqual(
            db_logger.get_all_task_pairings_for_worker(worker_id_1)[0], pair_1
        )
        self.assertEqual(
            len(db_logger.get_all_task_pairings_for_worker(worker_id_2)), 2
        )
        self.assertEqual(
            len(db_logger.get_all_task_pairings_for_worker(worker_id_1, 'fake_id')), 0
        )

        conversation_id_1 = "CONV_ID_1"
        conversation_id_2 = "CONV_ID_2"

        onboarding_id_1 = 'onboard_1'
        onboarding_id_2 = 'onboard_2'
        onboarding_id_3 = 'onboard_3'

        db_logger.log_start_onboard(worker_id_1, assignment_id_1, onboarding_id_1)
        db_logger.log_start_onboard(worker_id_2, assignment_id_2, onboarding_id_2)
        db_logger.log_start_onboard(worker_id_2, assignment_id_3, onboarding_id_3)
        db_logger.log_finish_onboard(worker_id_1, assignment_id_1)
        db_logger.log_finish_onboard(worker_id_2, assignment_id_2)
        db_logger.log_finish_onboard(worker_id_2, assignment_id_3)
        db_logger.log_start_task(worker_id_1, assignment_id_1, conversation_id_1)
        db_logger.log_start_task(worker_id_2, assignment_id_2, conversation_id_1)
        db_logger.log_start_task(worker_id_2, assignment_id_3, conversation_id_2)

        # Check to see retrieval by conversation
        pairs_1 = db_logger.get_pairings_for_conversation(conversation_id_1)
        pairs_2 = db_logger.get_pairings_for_conversation(conversation_id_2)
        pairs_3 = db_logger.get_pairings_for_conversation('fake_id')
        pair_1 = db_logger.get_worker_assignment_pairing(worker_id_1, assignment_id_1)
        pair_2 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_2)
        pair_3 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_3)
        self.assertEqual(pairs_1[0], pair_1)
        self.assertEqual(pairs_1[1], pair_2)
        self.assertEqual(pairs_2[0], pair_3)
        self.assertEqual(len(pairs_3), 0)

        # Do some final processing on assignments
        db_logger.log_complete_assignment(
            worker_id_1,
            assignment_id_1,
            time.time(),
            AssignState.STATUS_PARTNER_DISCONNECT,
        )
        db_logger.log_disconnect_assignment(
            worker_id_2, assignment_id_2, time.time(), AssignState.STATUS_DISCONNECT
        )
        db_logger.log_complete_assignment(
            worker_id_2, assignment_id_3, time.time(), AssignState.STATUS_DONE
        )

        # Assignment state consistent
        assignment_1_data = db_logger.get_assignment_data(assignment_id_1)
        assignment_2_data = db_logger.get_assignment_data(assignment_id_2)
        assignment_3_data = db_logger.get_assignment_data(assignment_id_3)
        self.assertEqual(assignment_1_data['assignment_id'], assignment_id_1)
        self.assertEqual(assignment_1_data['status'], 'Completed')
        self.assertIsNotNone(assignment_1_data['approve_time'])
        self.assertEqual(assignment_1_data['worker_id'], worker_id_1)
        self.assertEqual(assignment_1_data['hit_id'], HIT1['HIT']['HITId'])
        self.assertEqual(assignment_2_data['assignment_id'], assignment_id_2)
        self.assertEqual(assignment_2_data['status'], 'Disconnected')
        self.assertIsNotNone(assignment_2_data['approve_time'])
        self.assertEqual(assignment_2_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_2_data['hit_id'], HIT2['HIT']['HITId'])
        self.assertEqual(assignment_3_data['assignment_id'], assignment_id_3)
        self.assertEqual(assignment_3_data['status'], 'Completed')
        self.assertIsNotNone(assignment_3_data['approve_time'])
        self.assertEqual(assignment_3_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_3_data['hit_id'], HIT3['HIT']['HITId'])

        # Worker state consistent
        worker_1_data = db_logger.get_worker_data(worker_id_1)
        worker_2_data = db_logger.get_worker_data(worker_id_2)
        self.assertEqual(worker_1_data['worker_id'], worker_id_1)
        self.assertEqual(worker_1_data['accepted'], 1)
        self.assertEqual(worker_1_data['disconnected'], 0)
        self.assertEqual(worker_1_data['completed'], 1)
        self.assertEqual(worker_1_data['approved'], 0)
        self.assertEqual(worker_1_data['rejected'], 0)
        self.assertEqual(worker_2_data['worker_id'], worker_id_2)
        self.assertEqual(worker_2_data['accepted'], 2)
        self.assertEqual(worker_2_data['disconnected'], 1)
        self.assertEqual(worker_2_data['completed'], 1)
        self.assertEqual(worker_2_data['approved'], 0)
        self.assertEqual(worker_2_data['rejected'], 0)

        # Ensure Pairing state is consistent
        pair_1 = db_logger.get_worker_assignment_pairing(worker_id_1, assignment_id_1)
        pair_2 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_2)
        pair_3 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_3)
        self.assertEqual(pair_1['status'], AssignState.STATUS_PARTNER_DISCONNECT)
        self.assertEqual(pair_2['status'], AssignState.STATUS_DISCONNECT)
        self.assertEqual(pair_3['status'], AssignState.STATUS_DONE)
        self.assertEqual(pair_1['worker_id'], worker_id_1)
        self.assertEqual(pair_2['worker_id'], worker_id_2)
        self.assertEqual(pair_3['worker_id'], worker_id_2)
        self.assertEqual(pair_1['assignment_id'], assignment_id_1)
        self.assertEqual(pair_2['assignment_id'], assignment_id_2)
        self.assertEqual(pair_3['assignment_id'], assignment_id_3)
        self.assertEqual(pair_1['conversation_id'], conversation_id_1)
        self.assertEqual(pair_2['conversation_id'], conversation_id_1)
        self.assertEqual(pair_3['conversation_id'], conversation_id_2)
        self.assertGreaterEqual(pair_1['onboarding_end'], pair_1['onboarding_start'])
        self.assertGreaterEqual(pair_2['onboarding_end'], pair_2['onboarding_start'])
        self.assertGreaterEqual(pair_3['onboarding_end'], pair_3['onboarding_start'])
        self.assertGreaterEqual(pair_1['task_start'], pair_1['onboarding_end'])
        self.assertGreaterEqual(pair_2['task_start'], pair_2['onboarding_end'])
        self.assertGreaterEqual(pair_3['task_start'], pair_3['onboarding_end'])
        self.assertGreaterEqual(pair_1['task_end'], pair_1['onboarding_start'])
        self.assertGreaterEqual(pair_2['task_end'], pair_2['onboarding_start'])
        self.assertGreaterEqual(pair_3['task_end'], pair_3['onboarding_start'])
        self.assertEqual(pair_1['run_id'], run_id)
        self.assertEqual(pair_2['run_id'], run_id)
        self.assertEqual(pair_3['run_id'], run_id)
        self.assertEqual(pair_1['bonus_amount'], 0)
        self.assertEqual(pair_2['bonus_amount'], 0)
        self.assertEqual(pair_3['bonus_amount'], 0)
        self.assertEqual(pair_1['bonus_text'], '')
        self.assertEqual(pair_2['bonus_text'], '')
        self.assertEqual(pair_3['bonus_text'], '')
        self.assertFalse(pair_1['bonus_paid'])
        self.assertFalse(pair_2['bonus_paid'])
        self.assertFalse(pair_3['bonus_paid'])

        # Ensure run state is consistent
        run_data = db_logger.get_run_data(run_id)
        self.assertEqual(run_data['run_id'], run_id)
        self.assertEqual(run_data['created'], 3)
        self.assertEqual(run_data['completed'], 2)
        self.assertEqual(run_data['maximum'], hits_created)
        self.assertEqual(run_data['failed'], 1)

        # Test "submitting" and abandoning hits
        db_logger.log_submit_assignment(worker_id_1, assignment_id_1)
        db_logger.log_abandon_assignment(worker_id_2, assignment_id_2)
        db_logger.log_expire_assignment(worker_id_2, assignment_id_3)
        assignment_1_data = db_logger.get_assignment_data(assignment_id_1)
        assignment_2_data = db_logger.get_assignment_data(assignment_id_2)
        assignment_3_data = db_logger.get_assignment_data(assignment_id_3)
        self.assertEqual(assignment_1_data['assignment_id'], assignment_id_1)
        self.assertEqual(assignment_1_data['status'], 'Reviewable')
        self.assertIsNotNone(assignment_1_data['approve_time'])
        self.assertEqual(assignment_1_data['worker_id'], worker_id_1)
        self.assertEqual(assignment_1_data['hit_id'], HIT1['HIT']['HITId'])
        self.assertEqual(assignment_2_data['assignment_id'], assignment_id_2)
        self.assertEqual(assignment_2_data['status'], 'Abandoned')
        self.assertIsNotNone(assignment_2_data['approve_time'])
        self.assertEqual(assignment_2_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_2_data['hit_id'], HIT2['HIT']['HITId'])
        self.assertEqual(assignment_3_data['assignment_id'], assignment_id_3)
        self.assertEqual(assignment_3_data['status'], 'Expired')
        self.assertIsNotNone(assignment_3_data['approve_time'])
        self.assertEqual(assignment_3_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_3_data['hit_id'], HIT3['HIT']['HITId'])

        # Test approving and rejecting
        test_dollars = 3
        test_cents = 100 * test_dollars
        reason_use = 'Just because'
        out_reason = '${} for {}\n'.format(test_dollars, reason_use)
        db_logger.log_award_amount(
            worker_id_1, assignment_id_1, test_dollars, reason_use
        )
        db_logger.log_award_amount(
            worker_id_2, assignment_id_2, test_dollars, reason_use
        )
        db_logger.log_bonus_paid(worker_id_1, assignment_id_1)
        db_logger.log_approve_assignment(assignment_id_1)
        db_logger.log_reject_assignment(assignment_id_2)

        # Ensure state is valid again
        assignment_1_data = db_logger.get_assignment_data(assignment_id_1)
        assignment_2_data = db_logger.get_assignment_data(assignment_id_2)
        self.assertEqual(assignment_1_data['assignment_id'], assignment_id_1)
        self.assertEqual(assignment_1_data['status'], 'Approved')
        self.assertIsNotNone(assignment_1_data['approve_time'])
        self.assertEqual(assignment_1_data['worker_id'], worker_id_1)
        self.assertEqual(assignment_1_data['hit_id'], HIT1['HIT']['HITId'])
        self.assertEqual(assignment_2_data['assignment_id'], assignment_id_2)
        self.assertEqual(assignment_2_data['status'], 'Rejected')
        self.assertIsNotNone(assignment_2_data['approve_time'])
        self.assertEqual(assignment_2_data['worker_id'], worker_id_2)
        self.assertEqual(assignment_2_data['hit_id'], HIT2['HIT']['HITId'])

        worker_1_data = db_logger.get_worker_data(worker_id_1)
        worker_2_data = db_logger.get_worker_data(worker_id_2)
        self.assertEqual(worker_1_data['worker_id'], worker_id_1)
        self.assertEqual(worker_1_data['accepted'], 1)
        self.assertEqual(worker_1_data['disconnected'], 0)
        self.assertEqual(worker_1_data['expired'], 0)
        self.assertEqual(worker_1_data['completed'], 1)
        self.assertEqual(worker_1_data['approved'], 1)
        self.assertEqual(worker_1_data['rejected'], 0)
        self.assertEqual(worker_2_data['worker_id'], worker_id_2)
        self.assertEqual(worker_2_data['accepted'], 2)
        self.assertEqual(worker_2_data['disconnected'], 2)
        self.assertEqual(worker_2_data['expired'], 1)
        self.assertEqual(worker_2_data['completed'], 1)
        self.assertEqual(worker_2_data['approved'], 0)
        self.assertEqual(worker_2_data['rejected'], 1)

        pair_1 = db_logger.get_worker_assignment_pairing(worker_id_1, assignment_id_1)
        pair_2 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_2)
        pair_3 = db_logger.get_worker_assignment_pairing(worker_id_2, assignment_id_3)
        self.assertEqual(pair_1['status'], AssignState.STATUS_PARTNER_DISCONNECT)
        self.assertEqual(pair_2['status'], AssignState.STATUS_DISCONNECT)
        self.assertEqual(pair_3['status'], AssignState.STATUS_EXPIRED)
        self.assertEqual(pair_1['worker_id'], worker_id_1)
        self.assertEqual(pair_2['worker_id'], worker_id_2)
        self.assertEqual(pair_3['worker_id'], worker_id_2)
        self.assertEqual(pair_1['assignment_id'], assignment_id_1)
        self.assertEqual(pair_2['assignment_id'], assignment_id_2)
        self.assertEqual(pair_3['assignment_id'], assignment_id_3)
        self.assertEqual(pair_1['conversation_id'], conversation_id_1)
        self.assertEqual(pair_2['conversation_id'], conversation_id_1)
        self.assertEqual(pair_3['conversation_id'], conversation_id_2)
        self.assertGreaterEqual(pair_1['onboarding_end'], pair_1['onboarding_start'])
        self.assertGreaterEqual(pair_2['onboarding_end'], pair_2['onboarding_start'])
        self.assertGreaterEqual(pair_3['onboarding_end'], pair_3['onboarding_start'])
        self.assertGreaterEqual(pair_1['task_start'], pair_1['onboarding_end'])
        self.assertGreaterEqual(pair_2['task_start'], pair_2['onboarding_end'])
        self.assertGreaterEqual(pair_3['task_start'], pair_3['onboarding_end'])
        self.assertGreaterEqual(pair_1['task_end'], pair_1['onboarding_start'])
        self.assertGreaterEqual(pair_2['task_end'], pair_2['onboarding_start'])
        self.assertGreaterEqual(pair_3['task_end'], pair_3['onboarding_start'])
        self.assertEqual(pair_1['run_id'], run_id)
        self.assertEqual(pair_2['run_id'], run_id)
        self.assertEqual(pair_3['run_id'], run_id)
        self.assertEqual(pair_1['bonus_amount'], test_cents)
        self.assertEqual(pair_2['bonus_amount'], test_cents)
        self.assertEqual(pair_3['bonus_amount'], 0)
        self.assertEqual(pair_1['bonus_text'], out_reason)
        self.assertEqual(pair_2['bonus_text'], out_reason)
        self.assertEqual(pair_3['bonus_text'], '')
        self.assertTrue(pair_1['bonus_paid'])
        self.assertFalse(pair_2['bonus_paid'])
        self.assertFalse(pair_3['bonus_paid'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
This file contains a list of all the tasks, their id and task name, description and the
tags associated with them.
"""

task_list = [
    {
        "id": "AmazonQA",
        "display_name": "AmazonQA",
        "task": "amazon_qa",
        "tags": ["All", "QA"],
        "description": (
            "This dataset contains Question and Answer data from Amazon, "
            "totaling around 1.4 million answered questions."
            "Link: http://jmcauley.ucsd.edu/data/amazon/qa/"
        ),
    },
    {
        "id": "AQuA",
        "display_name": "AQuA",
        "task": "aqua",
        "tags": ["All", "QA"],
        "description": (
            "Dataset containing algebraic word problems with rationales for "
            "their answers. From Ling et. al. 2017, Link: "
            "https://arxiv.org/pdf/1705.04146.pdf"
        ),
    },
    {
        "id": "bAbI-1k",
        "display_name": "bAbI 1k",
        "task": "babi:All1k",
        "tags": ["All", "QA"],
        "description": (
            "20 synthetic tasks that each test a unique aspect of text and "
            "reasoning, and hence test different capabilities of learning "
            "models. From Weston et al. '16. Link: "
            "http://arxiv.org/abs/1502.05698 "
        ),
        "notes": (
            "You can access just one of the bAbI tasks with e.g. "
            "'babi:Task1k:3' for task 3."
        ),
    },
    {
        "id": "bAbI-10k",
        "display_name": "bAbI 10k",
        "task": "babi:All10k",
        "tags": ["All", "QA"],
        "description": (
            "20 synthetic tasks that each test a unique aspect of text and "
            "reasoning, and hence test different capabilities of learning "
            "models. From Weston et al. '16. Link: "
            "http://arxiv.org/abs/1502.05698"
        ),
        "notes": (
            "You can access just one of the bAbI tasks with e.g. 'babi:Task10k:3' "
            "for task 3."
        ),
    },
    {
        "id": "BlendedSkillTalk",
        "display_name": "Blended Skill Talk",
        "task": "blended_skill_talk",
        "tags": ["All", "ChitChat"],
        "description": (
            "A dataset of 7k conversations explicitly designed to exhibit multiple "
            "conversation modes: displaying personality, having empathy, and "
            "demonstrating knowledge."
        ),
    },
    {
        "id": "BookTest",
        "display_name": "BookTest",
        "task": "booktest",
        "tags": ["All", "Cloze"],
        "description": (
            "Sentence completion given a few sentences as context from a book. "
            "A larger version of CBT. From Bajgar et al., 16. Link: "
            "https://arxiv.org/abs/1610.00956"
        ),
    },
    {
        "id": "CBT",
        "display_name": "Children's Book Test (CBT)",
        "task": "cbt",
        "tags": ["All", "Cloze"],
        "description": (
            "Sentence completion given a few sentences as context from a "
            "children's book. From Hill et al., '16. Link: "
            "https://arxiv.org/abs/1511.02301"
        ),
    },
    {
        "id": "CCPE",
        "display_name": "Coached Conversational Preference Elicitation",
        "task": "ccpe",
        "tags": ["All", "Goal"],
        "description": (
            "A dataset consisting of 502 dialogs with 12,000 annotated "
            "utterances between a user and an assistant discussing movie "
            "preferences in natural language. It was collected using a "
            "Wizard-of-Oz methodology between two paid crowd-workers, "
            "where one worker plays the role of an 'assistant', while "
            "the other plays the role of a 'user'. From Google, '19. Link: "
            "https://ai.google/tools/datasets/coached-conversational-preference-elicitation"
        ),
    },
    {
        "id": "COPA",
        "display_name": "Choice of Plausible Alternatives",
        "task": "copa",
        "tags": ["All", "Reasoning"],
        "description": (
            "The Choice Of Plausible Alternatives (COPA) evaluation provides "
            "researchers with a tool for assessing progress in open-domain "
            "commonsense causal reasoning. COPA consists of 1000 questions, "
            "split equally into development and test sets of 500 questions "
            "each. See http://people.ict.usc.edu/~gordon/copa.html for more "
            "information"
        ),
    },
    {
        "id": "COQA",
        "display_name": "Conversational Question Answering Challenge",
        "task": "coqa",
        "tags": ["All", "QA"],
        "description": (
            "CoQA is a large-scale dataset for building Conversational "
            "Question Answering systems. The goal of the CoQA challenge "
            "is to measure the ability of machines to understand a text "
            "passage and answer a series of interconnected questions that "
            "appear in a conversation. CoQA is pronounced as coca . See "
            "https://arxiv.org/abs/1808.07042"
        ),
    },
    {
        "id": "CornellMovie",
        "display_name": "Cornell Movie",
        "task": "cornell_movie",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": (
            "Fictional conversations extracted from raw movie scripts. "
            "Danescu-Niculescu-Mizil & Lee, '11. Link: "
            "https://arxiv.org/abs/1106.3077"
        ),
    },
    {
        "id": "DBLL-bAbI",
        "display_name": "Dialog Based Language Learning: bAbI Task",
        "task": "dbll_babi",
        "tags": ["All", "Goal"],
        "description": (
            "Short dialogs based on the bAbI tasks, but in the form of a "
            "question from a teacher, the answer from the student, and finally a "
            "comment on the answer from the teacher. The aim is to find learning "
            "models that use the comments to improve. From Weston '16. Link: "
            "https://arxiv.org/abs/1604.06045. Tasks can be accessed with a "
            "format like: 'parlai display_data -t "
            "dbll_babi:task:2_p0.5' which specifies task 2, and policy with 0.5 "
            "answers correct, see the paper for more details of the tasks."
        ),
    },
    {
        "id": "DBLL-Movie",
        "display_name": "Dialog Based Language Learning: WikiMovies Task",
        "task": "dbll_movie",
        "tags": ["All", "Goal"],
        "description": (
            "Short dialogs based on WikiMovies, but in the form of a question "
            "from a teacher, the answer from the student, and finally a comment "
            "on the answer from the teacher. The aim is to find learning models "
            "that use the comments to improve. From Weston '16. Link: "
            "https://arxiv.org/abs/1604.06045"
        ),
    },
    {
        "id": "dialog-bAbI",
        "display_name": "Dialog bAbI",
        "task": "dialog_babi",
        "tags": ["All", "Goal"],
        "description": (
            "Simulated dialogs of restaurant booking, from Bordes et al. '16. "
            "Link: https://arxiv.org/abs/1605.07683"
        ),
    },
    {
        "id": "dialog-bAbI-plus",
        "display_name": "Dialog bAbI+",
        "task": "dialog_babi_plus",
        "tags": ["All", "Goal"],
        "description": (
            "bAbI+ is an extension of the bAbI Task 1 dialogues with everyday "
            "incremental dialogue phenomena (hesitations, restarts, and "
            "corrections) which model the disfluencies and communication "
            "problems in everyday spoken interaction in real-world environments. "
            "See https://www.researchgate.net/publication/319128941_Challenging_Neural_"
            "Dialogue_Models_with_Natural_Data_Memory_Networks_Fail_on_"
            "Incremental_Phenomena,http://aclweb.org/anthology/D17-1235"
        ),
    },
    {
        "id": "dialogue-nli",
        "display_name": "Dialogue NLI",
        "task": "dialogue_nli",
        "tags": ["All", "ChitChat", "NLI"],
        "description": (
            "Dialogue NLI is a dataset that addresses the issue of consistency in "
            "dialogue models. "
            "See: https://wellecks.github.io/dialogue_nli/"
        ),
    },
    {
        "id": "dstc7",
        "display_name": "DSTC7 subtrack 1 - ubuntu",
        "task": "dstc7",
        "tags": ["All", "ChitChat"],
        "description": (
            "DSTC7 is a competition which provided a dataset of dialogs very "
            "similar to the ubuntu dataset. In particular, the subtrack 1 "
            "consists in predicting the next utterance. "
            "See: https://arxiv.org/pdf/1901.03461.pdf"
        ),
    },
    {
        "id": "FVQA",
        "display_name": "FVQA",
        "task": "fvqa",
        "tags": ["All", "Visual"],
        "description": (
            "The FVQA, a VQA dataset which requires, and supports, much deeper "
            "reasoning. We extend a conventional visual question answering "
            "dataset, which contains image-question-answer triplets, through "
            "additional image-question-answer-supporting fact tuples. The "
            "supporting fact is represented as a structural triplet, such as "
            "<Cat,CapableOf,ClimbingTrees>.  Link: "
            "https://arxiv.org/abs/1606.05433"
        ),
    },
    {
        "id": "DealNoDeal",
        "display_name": "Deal or No Deal",
        "task": "dealnodeal",
        "tags": ["All", "Negotiation"],
        "description": (
            "End-to-end negotiation task which requires two agents to agree on "
            "how to divide a set of items, with each agent assigning different "
            "values to each item. From Lewis et al. '17. Link: "
            "https://arxiv.org/abs/1706.05125"
        ),
    },
    {
        "id": "HotpotQA",
        "display_name": "HotpotQA",
        "task": "hotpotqa",
        "tags": ["All", "QA"],
        "description": (
            "HotpotQA is a dataset for multi-hop question answering."
            "The overall setting is that given some context paragraphs"
            "(e.g., a few paragraphs, or the entire Web) and a question,"
            "a QA system answers the question by extracting a span of text"
            "from the context. It is necessary to perform multi-hop reasoning"
            "to correctly answer the question."
            "Link: https://arxiv.org/pdf/1809.09600.pdf"
        ),
    },
    {
        "id": "LIGHT-Dialogue",
        "display_name": "LIGHT-Dialogue",
        "task": "light_dialog",
        "tags": ["All", "Grounded", "Dodeca"],
        "description": (
            "LIGHT is a text adventure game with actions and dialogue collected."
            "The source data is collected between crowdworkers playing the game."
            "Link: http://parl.ai/projects/light"
        ),
    },
    {
        "id": "MutualFriends",
        "display_name": "MutualFriends",
        "task": "mutualfriends",
        "tags": ["All", "Goal"],
        "description": (
            "Task where two agents must discover which friend of theirs is "
            "mutual based on the friends's attributes. From He He et al. '17. "
            "Link: https://stanfordnlp.github.io/cocoa/"
        ),
    },
    {
        "id": "MCTest",
        "display_name": "MCTest",
        "task": "mctest",
        "tags": ["All", "QA"],
        "description": (
            "Questions about short children's stories, from Richardson et al. "
            "'13. Link: https://www.microsoft.com/en-us/research/publication/"
            "mctest-challenge-dataset-open-domain-machine-comprehension-text/"
        ),
    },
    {
        "id": "MovieDD-QA",
        "display_name": "Movie Dialog QA",
        "task": "moviedialog:Task:1",
        "tags": ["All", "QA", "MovieDD"],
        "description": (
            "Closed-domain QA dataset asking templated questions about movies, "
            "answerable from Wikipedia, similar to WikiMovies. From Dodge et al. "
            "'15. Link: https://arxiv.org/abs/1511.06931"
        ),
    },
    {
        "id": "MovieDD-QARecs",
        "display_name": "Movie Dialog QA Recommendations",
        "task": "moviedialog:Task:3",
        "tags": ["All", "Goal", "MovieDD"],
        "description": (
            "Dialogs discussing questions about movies as well as "
            "recommendations. From Dodge et al. '15. Link: "
            "https://arxiv.org/abs/1511.06931"
        ),
    },
    {
        "id": "MovieDD-Recs",
        "display_name": "Movie Dialog Recommendations",
        "task": "moviedialog:Task:2",
        "tags": ["All", "QA", "MovieDD"],
        "description": (
            "Questions asking for movie recommendations. From Dodge et al. '15. "
            "Link: https://arxiv.org/abs/1511.06931"
        ),
    },
    {
        "id": "MovieDD-Reddit",
        "display_name": "Movie Dialog Reddit",
        "task": "moviedialog:Task:4",
        "tags": ["All", "ChitChat", "MovieDD"],
        "description": (
            "Dialogs discussing Movies from Reddit (the Movies SubReddit). From "
            "Dodge et al. '15. Link: https://arxiv.org/abs/1511.06931"
        ),
    },
    {
        "id": "MTurkWikiMovies",
        "display_name": "MTurk WikiMovies",
        "task": "mturkwikimovies",
        "tags": ["All", "QA"],
        "description": (
            "Closed-domain QA dataset asking MTurk-derived questions about "
            "movies, answerable from Wikipedia. From Li et al. '16. Link: "
            "https://arxiv.org/abs/1611.09823"
        ),
    },
    {
        "id": "MultiNLI",
        "display_name": "MultiNLI",
        "task": "multinli",
        "tags": ["All", "Entailment", "decanlp"],
        "description": (
            "A dataset designed for use in the development and evaluation of "
            "machine learning models for sentence understanding. Each example "
            "contains a premise and hypothesis. Model has to predict whether "
            "premise and hypothesis entail, contradict or are neutral to each "
            "other. From Williams et al. '17. Link: "
            "https://arxiv.org/abs/1704.05426"
        ),
    },
    {
        "id": "NarrativeQA",
        "display_name": "NarrativeQA",
        "task": "narrative_qa",
        "tags": ["All", "QA"],
        "description": (
            "A dataset and set of tasks in which the reader must answer "
            "questions about stories by reading entire books or movie scripts. "
            "From Kočiský et. al. '17. Link: https://arxiv.org/abs/1712.07040'"
        ),
        "notes": (
            "You can access summaries only task for NarrativeQA by using task "
            "'narrative_qa:summaries'. By default, only stories are provided."
        ),
    },
    {
        "id": "OpenSubtitles",
        "display_name": "Open Subtitles",
        "task": "opensubtitles",
        "tags": ["All", "ChitChat"],
        "description": (
            "Dataset of dialogs from movie scripts. Version 2018: "
            "http://opus.lingfil.uu.se/OpenSubtitles2018.php, version 2009: "
            "http://opus.lingfil.uu.se/OpenSubtitles.php. A variant of the "
            "dataset used in Vinyals & Le '15, "
            "https://arxiv.org/abs/1506.05869."
        ),
    },
    {
        "id": "personalized-dialog-full",
        "display_name": "Personalized Dialog Full Set",
        "task": "personalized_dialog:AllFull",
        "tags": ["All", "Goal", "Personalization"],
        "description": (
            "Simulated dataset of restaurant booking focused on personalization "
            "based on user profiles. From Joshi et al. '17. Link: "
            "https://arxiv.org/abs/1706.07503"
        ),
    },
    {
        "id": "personalized-dialog-small",
        "display_name": "Personalized Dialog Small Set",
        "task": "personalized_dialog:AllSmall",
        "tags": ["All", "Goal", "Personalization"],
        "description": (
            "Simulated dataset of restaurant booking focused on personalization "
            "based on user profiles. From Joshi et al. '17. Link: "
            "https://arxiv.org/abs/1706.07503"
        ),
    },
    {
        "id": "QACNN",
        "display_name": "QA CNN",
        "task": "qacnn",
        "tags": ["All", "Cloze"],
        "description": (
            "Cloze dataset based on a missing (anonymized) entity phrase from a "
            "CNN article, Hermann et al. '15. Link: "
            "https://arxiv.org/abs/1506.03340"
        ),
    },
    {
        "id": "QADailyMail",
        "display_name": "QA Daily Mail",
        "task": "qadailymail",
        "tags": ["All", "Cloze"],
        "description": (
            "Cloze dataset based on a missing (anonymized) entity phrase from a "
            "Daily Mail article, Hermann et al. '15. Link: "
            "https://arxiv.org/abs/1506.03340"
        ),
    },
    {
        "id": "QuAC",
        "display_name": "Question Answering in Context",
        "task": "quac",
        "tags": ["All", "QA"],
        "description": (
            "Question Answering in Context is a dataset for modeling, "
            "understanding, and participating in information seeking dialog. Data "
            "instances consist of an interactive dialog between two crowd workers: "
            "(1) a student who poses a sequence of freeform questions to learn as "
            "much as possible about a hidden Wikipedia text, and (2) a teacher who "
            "answers the questions by providing short excerpts (spans) from the text. "
            "QuAC introduces challenges not found in existing machine comprehension "
            "datasets: its questions are often more open-ended, unanswerable, "
            "or only meaningful within the dialog context. link: "
            "https://arxiv.org/abs/1808.07036"
        ),
    },
    {
        "id": "SelfFeedingChatbot",
        "display_name": "Self-Feeding Chatbot",
        "task": "self_feeding",
        "tags": ["diaexp", "diasen", "All"],
        "description": (
            "Learning from Dialogue after Deployment. Leveraging user textual "
            "feedback to improve the chatbot's abilities. "
            "From Hancock et al. 2019, Link: "
            "https://arxiv.org/abs/1901.05415"
        ),
    },
    {
        "id": "SimpleQuestions",
        "display_name": "Simple Questions",
        "task": "simplequestions",
        "tags": ["All", "QA"],
        "description": (
            "Open-domain QA dataset based on Freebase triples from Bordes et "
            "al. '15. Link: https://arxiv.org/abs/1506.02075"
        ),
    },
    {
        "id": "SNLI",
        "display_name": "The Stanford Natural Language Inference (SNLI) Corpus",
        "task": "snli",
        "tags": ["All", "Entailment"],
        "description": (
            "The SNLI corpus (version 1.0) is a collection of 570k "
            "human-written English sentence pairs manually labeled for balanced "
            "classification with the labels entailment, contradiction, and "
            "neutral, supporting the task of natural language inference (NLI), "
            "also known as recognizing textual entailment (RTE). See "
            "https://nlp.stanford.edu/projects/snli/"
        ),
    },
    {
        "id": "SQuAD2",
        "display_name": "SQuAD2",
        "task": "squad2",
        "tags": ["All", "QA"],
        "description": (
            "Open-domain QA dataset answerable from a given paragraph from "
            "Wikipedia, from Rajpurkar & Jia et al. '18. Link: "
            "http://arxiv.org/abs/1806.03822"
        ),
    },
    {
        "id": "SQuAD",
        "display_name": "SQuAD",
        "task": "squad",
        "tags": ["All", "QA"],
        "description": (
            "Open-domain QA dataset answerable from a given paragraph from "
            "Wikipedia, from Rajpurkar et al. '16. Link: "
            "https://arxiv.org/abs/1606.05250"
        ),
    },
    {
        "id": "TriviaQA",
        "display_name": "TriviaQA",
        "task": "triviaqa",
        "tags": ["All", "QA"],
        "description": (
            "Open-domain QA dataset with question-answer-evidence triples, from "
            "Joshi et al. '17. Link: https://arxiv.org/abs/1705.03551"
        ),
    },
    {
        "id": "TaskNTalk",
        "display_name": "Task N' Talk",
        "task": "taskntalk",
        "tags": ["All", "Goal"],
        "description": (
            "Dataset of synthetic shapes described by attributes, for agents to "
            "play a cooperative QA game, from Kottur et al. '17. Link: "
            "https://arxiv.org/abs/1706.08502"
        ),
    },
    {
        "id": "Ubuntu",
        "display_name": "Ubuntu",
        "task": "ubuntu",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": (
            "Dialogs between an Ubuntu user and an expert trying to fix issue, "
            "we use the V2 version, which cleaned the data to some extent. "
            "From Lowe et al. '15. Link: https://arxiv.org/abs/1506.08909."
        ),
    },
    {
        "id": "WebQuestions",
        "display_name": "Web Questions",
        "task": "webquestions",
        "tags": ["All", "QA"],
        "description": (
            "Open-domain QA dataset from Web queries from Berant et al. '13. "
            "Link: http://www.aclweb.org/anthology/D13-1160"
        ),
    },
    {
        "id": "WikiMovies",
        "display_name": "WikiMovies",
        "task": "wikimovies",
        "tags": ["All", "QA"],
        "description": (
            "Closed-domain QA dataset asking templated questions about movies, "
            "answerable from Wikipedia. From Miller et al. '16. Link: "
            "https://arxiv.org/abs/1606.03126"
        ),
    },
    {
        "id": "WikiQA",
        "display_name": "WikiQA",
        "task": "wikiqa",
        "tags": ["All", "QA"],
        "description": (
            "Open domain QA from Wikipedia dataset from Yang et al. '15. Link: "
            "https://www.microsoft.com/en-us/research/publication/wikiqa-a-"
            "challenge-dataset-for-open-domain-question-answering/"
        ),
    },
    {
        "id": "VQAv1",
        "display_name": "VQAv1",
        "task": "vqa_v1",
        "tags": ["All", "Visual"],
        "description": (
            "Open-ended question answering about visual content. From Agrawal "
            "et al. '15. Link: https://arxiv.org/abs/1505.00468"
        ),
    },
    {
        "id": "VQAv2",
        "display_name": "VQAv2",
        "task": "vqa_v2",
        "tags": ["All", "Visual"],
        "description": (
            "Bigger, more balanced version of the original VQA dataset. From "
            "Goyal et al. '16. Link: https://arxiv.org/abs/1612.00837"
        ),
    },
    {
        "id": "VisDial",
        "display_name": "VisDial",
        "task": "visdial",
        "tags": ["All", "Visual"],
        "description": (
            "Task which requires agents to hold a meaningful dialog about "
            "visual content. From Das et al. '16. Link: "
            "https://arxiv.org/abs/1611.08669"
        ),
    },
    {
        "id": "MNIST_QA",
        "display_name": "MNIST_QA",
        "task": "mnist_qa",
        "tags": ["All", "Visual"],
        "description": (
            "Task which requires agents to identify which number they are "
            "seeing. From the MNIST dataset."
        ),
    },
    {
        "id": "InsuranceQA",
        "display_name": "InsuranceQA",
        "task": "insuranceqa",
        "tags": ["All", "QA"],
        "description": (
            "Task which requires agents to identify high quality answers "
            "composed by professionals with deep domain knowledge. From Feng et "
            "al. '15. Link: https://arxiv.org/abs/1508.01585"
        ),
    },
    {
        "id": "MS_MARCO",
        "display_name": "MS_MARCO",
        "task": "ms_marco",
        "tags": ["All", "QA"],
        "description": (
            "A large scale Machine Reading Comprehension Dataset with questions "
            "sampled from real anonymized user queries and contexts from web "
            "documents. From Nguyen et al. '16. Link: "
            "https://arxiv.org/abs/1611.09268"
        ),
    },
    {
        "id": "CLEVR",
        "display_name": "CLEVR",
        "task": "clevr",
        "tags": ["All", "Visual"],
        "description": (
            "A visual reasoning dataset that tests abilities such as attribute "
            "identification, counting, comparison, spatial relationships, and "
            "logical operations. From Johnson et al. '16. Link: "
            "https://arxiv.org/abs/1612.06890"
        ),
    },
    {
        "id": "nlvr",
        "display_name": "nlvr",
        "task": "nlvr",
        "tags": ["All", "Visual"],
        "description": (
            "Cornell Natural Language Visual Reasoning (NLVR) is a language "
            "grounding dataset based on  pairs of natural language statements "
            "grounded in synthetic images. From Suhr et al. '17. Link: "
            "http://lic.nlp.cornell.edu/nlvr/"
        ),
    },
    {
        "id": "WMT",
        "display_name": "WMT",
        "task": "wmt",
        "tags": ["All", "MT"],
        "description": (
            "Workshop on Machine Translation task, currently only includes en_de."
        ),
    },
    {
        "id": "IWSLT14",
        "display_name": "IWSLT14",
        "task": "iwslt14",
        "tags": ["All", "MT", "decanlp"],
        "description": (
            "2014 International Workshop on Spoken Language task, currently "
            "only includes en_de and de_en. From Cettolo et al. '12. Link: "
            "wit3.fbk.eu"
        ),
    },
    {
        "id": "ConvAI2",
        "display_name": "ConvAI2",
        "task": "convai2",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": (
            "A chit-chat dataset based on PersonaChat "
            "(https://arxiv.org/abs/1801.07243) for a NIPS 2018 competition. "
            "Link: http://convai.io/."
        ),
    },
    {
        "id": "ConvAI_ChitChat",
        "display_name": "ConvAI_ChitChat",
        "task": "convai_chitchat",
        "tags": ["All", "ChitChat", "decanlp"],
        "description": (
            "Human-bot dialogues containing free discussions of randomly chosen "
            "paragraphs from SQuAD. Link to dataset: http://convai.io/data/"
        ),
    },
    {
        "id": "Dialogue_QE",
        "display_name": "Dialogue_QE",
        "task": "dialogue_qe",
        "tags": ["All"],
        "description": (
            "Human-bot dialogues labelled for quality at the level of "
            "dialogues. Can be used to train dialogue-level metric for dialogue "
            "systems. Link to dataset: http://convai.io/data/"
        ),
    },
    {
        "id": "QAngaroo",
        "display_name": "QAngaroo",
        "task": "qangaroo",
        "tags": ["All", "QA"],
        "description": (
            "Reading Comprehension with Multiple Hop. Including two datasets: "
            "WIKIHOP built on on wikipedia, MEDHOP built on paper abstracts from "
            "PubMed. Link to dataset: http://qangaroo.cs.ucl.ac.uk/"
        ),
    },
    {
        "id": "SCAN",
        "display_name": "SCAN",
        "task": "scan",
        "tags": ["Goal", "All"],
        "description": (
            "SCAN is a set of simple language-driven navigation tasks for "
            "studying compositional learning and zero-shot generalization. The "
            "SCAN tasks were inspired by the CommAI environment, which is the "
            "origin of the acronym (Simplified versions of the CommAI Navigation "
            "tasks). See the paper: https://arxiv.org/abs/1711.00350 or data: "
            "https://github.com/brendenlake/SCAN"
        ),
    },
    {
        "id": "Persona-Chat",
        "display_name": "Persona-Chat",
        "task": "personachat",
        "tags": ["ChitChat", "All"],
        "description": (
            "A chit-chat dataset where paired Turkers are given assigned "
            "personas and chat to try to get to know each other. See the paper: "
            "https://arxiv.org/abs/1801.07243"
        ),
    },
    {
        "id": "TaskMaster",
        "display_name": "TaskMaster-1-2019",
        "task": "taskmaster",
        "tags": ["ChitChat", "All"],
        "description": (
            "A chit-chat dataset by GoogleAI providing high quality goal-oriented conversations"
            "The dataset hopes to provoke interest in written vs spoken language"
            "Both the datasets consists of two-person dialogs:"
            "Spoken: Created using Wizard of Oz methodology.(woz-dialogs.json)"
            "Written: Created by crowdsourced workers who were asked to write the full conversation themselves playing roles of both the user and assistant. (self-dialogs.json)"
            "Link: https://ai.google/tools/datasets/taskmaster-1"
        ),
    },
    {
        "id": "Twitter",
        "display_name": "Twitter",
        "task": "twitter",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": (
            "Twitter data from: https://github.com/Marsan-Ma/chat_corpus/. No "
            "train/valid/test split was provided so 10k for valid and 10k for "
            "test was chosen at random."
        ),
    },
    {
        "id": "Wikipedia",
        "display_name": "Wikipedia",
        "task": 'wikipedia',
        "tags": ["All"],
        "description": ("Dump of Wikipedia articles from 2/3/18"),
        "notes": (
            "Specify ':full' for the full articles to be returned, otherwise "
            "defaults to ':summary', which provides the first paragraphs. To put "
            "the article in the labels and the title in the text, specify "
            "':key-value' at the end (for a title/content key-value "
            "association)"
        ),
    },
    {
        "id": "Flickr30k",
        "display_name": "Flickr30k",
        "task": "flickr30k",
        "tags": ["All", "Visual"],
        "description": (
            "30k captioned images pulled from Flickr compiled by UIUC: "
            "http://web.engr.illinois.edu/~bplumme2/Flickr30kEntities/. Based "
            "off of these papers: https://arxiv.org/abs/1505.04870v2, "
            "http://aclweb.org/anthology/Q14-1006"
        ),
    },
    {
        "id": "COCO_Captions",
        "display_name": "COCO_Captions",
        "task": "coco_caption",
        "tags": ["All", "Visual"],
        "description": (
            "COCO annotations derived from the 2015 COCO Caption Competition. "
            "Link to dataset: http://cocodataset.org/#download"
        ),
    },
    {
        "id": "integration_tests",
        "display_name": "Integration Tests",
        "task": "integration_tests",
        "tags": ["All", "Debug"],
        "description": ("Artificial tasks for ensuring models perform as expected"),
    },
    {
        "id": "ConvAI2_wild_evaluation",
        "display_name": "ConvAI2_wild_evaluation",
        "task": "convai2_wild_evaluation",
        "tags": ["All", "ChitChat"],
        "description": (
            "Dataset collected during the wild evaluation of ConvaAI2 participants "
            "bots (http://convai.io). 60% train, 20% valid and 20% test is chosen at "
            "random from the whole dataset."
        ),
    },
    {
        "id": "sst",
        "display_name": "SST Sentiment Analysis",
        "task": "sst",
        "tags": ["All", "decanlp"],
        "description": (
            "Dataset containing sentiment trees of movie reviews. We use the modified "
            "binary sentence analysis subtask given by the DecaNLP paper here, "
            "originally from Radford, et. al "
            "https://nlp.stanford.edu/sentiment/index.html "
            "https://github.com/openai/generating-reviews-discovering-sentiment/"
        ),
    },
    {
        "id": "cnn_dm",
        "display_name": "CNN/DM Summarisation",
        "task": "cnn_dm",
        "tags": ["All", "decanlp"],
        "description": (
            "Dataset collected from CNN and the Daily Mail with summaries as labels, "
            "Implemented as part of the DecaNLP task"
            "Downloaded from https://cs.nyu.edu/~kcho/DMQA/"
        ),
    },
    {
        "id": "qasrl",
        "display_name": "QA-SRL Semantic Role Labeling",
        "task": "qasrl",
        "tags": ["All", "decanlp"],
        "description": (
            "QA dataset implemented as part of the DecaNLP task. More info on the"
            "dataset can be found here: https://dada.cs.washington.edu/qasrl/"
        ),
    },
    {
        "id": "qazre",
        "display_name": "QA-ZRE Relation Extraction",
        "task": "qazre",
        "tags": ["All", "decanlp"],
        "description": (
            "Zero Shot relation extraction task implemented as part of the DecaNLP "
            "task. More info on the dataset can be found here:"
            "http://nlp.cs.washington.edu/zeroshot/"
        ),
    },
    {
        "id": "woz",
        "display_name": "WOZ restuarant reservation (Goal-Oriented Dialogue)",
        "task": "woz",
        "tags": ["All", "decanlp"],
        "description": (
            "Dataset containing dialogues dengotiating a resturant reservation. "
            "Implemented as part of the DecaNLP task, focused on the change "
            "in the dialogue state. Original paper: "
            "https://arxiv.org/abs/1604.04562"
        ),
    },
    {
        "id": "wikisql",
        "display_name": "WikiSQL semantic parsing task",
        "task": "wikisql",
        "tags": ["All", "decanlp"],
        "description": (
            "Dataset for parsing sentences to SQL code, given a table. "
            "Implemented as part of the DecaNLP task. More info can be found here:"
            "https://github.com/salesforce/WikiSQL"
        ),
    },
    {
        "id": "mwsc",
        "display_name": "MWSC pronoun resolution",
        "task": "mwsc",
        "tags": ["All", "decanlp"],
        "description": "Resolving possible ambiguous pronouns. Implemented as part of the DecaNLP"
        "task, and can be found on the decaNLP github",
    },
    {
        "id": "decanlp",
        "display_name": "DecaNLP: The Natural Language Decathlon",
        "task": "decanlp",
        "tags": ["All"],
        "description": (
            "A collection of 10 tasks (SQuAD, IWSLT, CNN/DM, MNLI, SST, QA‑SRL,"
            "QA‑ZRE, WOZ, WikiSQL and MWSC) designed to challenge a model with a range "
            "of different tasks. Note that we use IWSLT 2014 instead of "
            "2016/2013test/2014test for train/dev/test as given in the DecaNLP paper. "
            "See paper https://arxiv.org/abs/1806.08730 for more information and "
            "github: https://github.com/salesforce/decaNLP for data sources"
        ),
    },
    {
        "id": "Personality_Captions",
        "display_name": "Personality_Captions",
        "task": "personality_captions",
        "tags": ["All", "Visual"],
        "description": (
            "200k images from the YFCC100m dataset "
            "(https://multimediacommons.wordpress.com/yfcc100m-core-dataset/), "
            "with captions conditioned on one of 215 personalities. See "
            "https://arxiv.org/abs/1810.10665 for more information."
        ),
        "notes": (
            "If you have already downloaded the images, please specify with "
            "the `--yfcc-path` flag, as the image download script takes a "
            "very long time to run"
        ),
    },
    {
        "id": "Image_Chat",
        "display_name": "Image_Chat",
        "task": "image_chat",
        "tags": ["All", "Visual", "ChitChat"],
        "description": (
            "202k dialogues and 401k utterances over 202k images from "
            "the YFCC100m dataset"
            "(https://multimediacommons.wordpress.com/yfcc100m-core-dataset/)"
            "using 215 possible personality traits"
            "see https://klshuster.github.io/image_chat/ for more information."
        ),
        "notes": (
            "If you have already downloaded the images, please specify with "
            "the `--yfcc-path` flag, as the image download script takes a "
            "very long time to run"
        ),
    },
    {
        "id": "Image_Chat_Generation",
        "display_name": "Image_Chat_Generation",
        "task": "image_chat:Generation",
        "tags": ["All", "Visual", "ChitChat", "Dodeca"],
        "description": ("Image Chat task to train generative model"),
    },
    {
        "id": "Wizard_of_Wikipedia",
        "display_name": "Wizard_of_Wikipedia",
        "task": "wizard_of_wikipedia",
        "tags": ["All", "ChitChat"],
        "description": (
            "A dataset with conversations directly grounded with knowledge "
            "retrieved from Wikipedia. Contains 201k utterances from 22k "
            "dialogues spanning over 1300 diverse topics, split into train, "
            "test, and valid sets. The test and valid sets are split "
            "into two sets each: one with overlapping topics with the train "
            "set, and one with unseen topics."
            "See https://arxiv.org/abs/1811.01241 for more information."
        ),
        "notes": (
            "To access the different valid/test splits (unseen/seen), specify "
            "the corresponding split (`random_split` for seen, `topic_split` "
            "for unseen) after the last colon in the task. "
            "E.g. `wizard_of_wikipedia:WizardDialogKnowledgeTeacher:random_split`"
        ),
    },
    {
        "id": "Wizard_of_Wikipedia_Generator",
        "display_name": "Wizard_of_Wikipedia_Generator",
        "task": "wizard_of_wikipedia:Generator",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": ("Wizard of Wikipedia task to train generative models"),
    },
    {
        "id": "DailyDialog",
        "display_name": "Daily Dialog",
        "task": "dailydialog",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": (
            "A dataset of chitchat dialogues with strong annotations for "
            "topic, emotion and utterance act. This version contains both sides "
            "of every conversation, and uses the official train/valid/test splits "
            "from the original authors. See https://arxiv.org/abs/1710.03957 "
            "for more information."
        ),
    },
    {
        "id": "EmpatheticDialogues",
        "display_name": "Empathetic Dialogues",
        "task": "empathetic_dialogues",
        "tags": ["All", "ChitChat", "Dodeca"],
        "description": (
            "A dataset of 25k conversations grounded in emotional situations "
            "to facilitate training and evaluating dialogue systems. See "
            "https://arxiv.org/abs/1811.00207 for more information. \n"
            "Dataset has been released under the CC BY-NC license. \n"
            "EmpatheticDialoguesTeacher returns examples like so: \n\n"
            "  - [text]:  context line (previous utterance by 'speaker') \n"
            "  - [labels]: label line  (current utterance by 'listener') \n\n"
            "with additional task specific fields: \n\n"
            "  - [situation]: a 1-3 sentence description of the situation that the conversation is \n"
            "  - [emotion]: one of 32 emotion words \n\n"
            "Other optional fields: \n\n"
            "  - [prepend_ctx]: fasttext prediction on context line - or None \n"
            "  - [prepend_cand]: fasttext prediction on label line (candidate) - or None \n"
            "  - [deepmoji_ctx]: vector encoding from deepmoji penultimate layer - or None \n"
            "  - [deepmoji_cand]: vector encoding from deepmoji penultimate layer for label line (candidate) - or None "
        ),
    },
    {
        "id": "DialogueSafety",
        "display_name": "Dialogue Safety",
        "task": "dialogue_safety",
        "tags": ["All"],
        "description": (
            "Several datasets described in the paper Built it Break it Fix it "
            "for Dialogue Safety: Robustness from Adversarial Human Attack "
            "(see https://arxiv.org/abs/1908.06083 for more information). \n"
            "All datasets are classification tasks in which the goal is to "
            "determine if the text is offensive or \'safe\'."
        ),
    },
    {
        "id": "MultiWOZ",
        "display_name": "MultiWOZ",
        "task": "multiwoz",
        "tags": ["All", "Goal"],
        "description": (
            "A fully labeled collection of human-written conversations spanning"
            "over multiple domains and topics."
            "(see http://dialogue.mi.eng.cam.ac.uk/index.php/corpus/"
            " for more information). "
        ),
    },
    {
        "id": "SelfChat",
        "display_name": "SelfChat",
        "task": "self_chat",
        "tags": [],
        "description": (
            "Not a dataset, but a generic world for model self-chats. "
            "(see parlai/scripts/self_chat.py"
            " for more information). "
        ),
    },
    {
        "id": "OneCommon",
        "display_name": "OneCommon",
        "task": "onecommon",
        "tags": ["All", "Goal"],
        "description": (
            "A collaborative referring task which requires advanced skills "
            "of common grounding under continuous and partially-observable context. "
            "This code also includes reference-resolution annotation "
            "from Udagawa and Aizawa '19. Link: https://github.com/Alab-NII/onecommon"
        ),
    },
    {
        "id": "IGC",
        "display_name": "Image Grounded Conversations",
        "task": "igc",
        "tags": ["All", "Visual", "ChitChat", "Dodeca"],
        "description": (
            "A dataset of (image, context, question, answer) tuples, comprised "
            "of eventful images taken from Bing, Flickr, and COCO. See "
            "https://arxiv.org/abs/1701.08251 for more information."
        ),
    },
    {
        "id": "ANLI",
        "display_name": "Adversarial Natural Language Inference (ANLI) Corpus",
        "task": "anli",
        "tags": ["All", "Entailment", "NLI"],
        "description": (
            "The ANLI corpus (version 1.0) is a new large-scale NLI benchmark dataset,"
            "collected via an iterative, adversarial human-and-model-in-the-loop procedure"
            "with the labels entailment, contradiction, and neutral. A total of three rounds "
            "of data are collected that progressively increase in difficulty and complexity."
            "See https://github.com/facebookresearch/anli and https://arxiv.org/abs/1910.14599 "
            "for more information."
        ),
    },
    {
        "id": "NLI",
        "display_name": "Natural Language Inference (NLI) Corpus",
        "task": "nli",
        "tags": ["All", "Entailment"],
        "description": (
            "A collection of 3 popular Natural Language Inference(NLI) benchmark tasks: "
            "ANLI v0.1, MultiNLI 1.0, SNLI 1.0."
        ),
    },
    {
        "id": "Funpedia",
        "display_name": "Funpedia",
        "task": "funpedia",
        "tags": ["All"],
        "description": (
            "Task for rephrasing sentences from Wikipedia conditioned on a persona."
        ),
    },
    {
        "id": "LIGHTGenderBias",
        "display_name": "LIGHT Gender Bias",
        "task": "light_genderation_bias",
        "tags": ["All"],
        "description": (
            "Task for debiasing the LIGHT dataset, all mitigation methods described here: "
            "<https://arxiv.org/abs/1911.03842>."
        ),
    },
    {
        "id": "AirDialogue",
        "display_name": "AirDialogue",
        "task": "airdialogue",
        "tags": ["All", "Goal"],
        "description": (
            "Task for goal-oriented dialogue using airplane booking conversations "
            "between agents and customers. Paper and toolkits for the dataset can be"
            " found at <https://github.com/google/airdialogue>."
        ),
    },
    {
        "id": "HollE",
        "display_name": "Holl-E",
        "task": "holl_e",
        "tags": ["All", "ChitChat"],
        "description": (
            "Sequence of utterances and responses with background knowledge about"
            "movies. From the Holl-E dataset. More information found at"
            " https://github.com/nikitacs16/Holl-E."
        ),
    },
    {
        "id": "ELI5",
        "display_name": "ELI5",
        "task": "eli5",
        "tags": ["All", "QA"],
        "description": (
            "This dataset contains Question and Answer data from Reddit "
            "explainlikeimfive posts and comments."
            "Link: https://github.com/facebookresearch/ELI5/"
        ),
    },
    {
        "id": "ReDial",
        "display_name": "ReDial",
        "task": "redial",
        "tags": ["All", "ChitChat", "Goal"],
        "description": (
            "Annotated dataset of dialogues where users recommend movies to eachother."
            "See https://redialdata.github.io/website/ for more information."
        ),
    },
    {
        "id": "DREAM",
        "display_name": "DREAM",
        "task": "dream",
        "tags": ["All", "QA"],
        "description": (
            "A multiple-choice answering dataset based on multi-turn, multi-party dialogue."
            "More information can be found at: <https://dataset.org/dream/>."
        ),
    },
    {
        "id": "C3",
        "display_name": "C3",
        "task": "c3",
        "tags": ["All", "QA"],
        "description": (
            "A multiple-choice answering dataset in Chinese based on a prior passage."
            "More information can be found at: <https://dataset.org/c3/>."
        ),
    },
    {
        "id": "CommonSenseQA",
        "display_name": "CommonSenseQA",
        "task": "commonsenseqa",
        "tags": ["All", "QA"],
        "description": (
            "CommonSenseQA is a multiple-choice Q-A dataset that relies on commonsense "
            "knowlegde to predict correct answers. More information found at "
            "<https://www.tau-nlp.org/commonsenseqa>."
        ),
    },
    {
        "id": "StyleGen",
        "display_name": "Style-Controlled Generation",
        "task": "style_gen",
        "tags": ["All", "ChitChat"],
        "description": (
            "Dialogue datasets (BlendedSkillTalk, ConvAI2, EmpatheticDialogues, and "
            "Wizard of Wikipedia) labeled with personalities taken from the Image-Chat "
            "dataset. Used for the style-controlled generation project"
        ),
    },
    {
        "id": "GoogleSGD",
        "display_name": "GoogleSGD",
        "task": "google_sgd",
        "tags": ["All", "Goal"],
        "description": (
            "The Schema-Guided Dialogue (SGD) dataset consists of over 20k "
            "annotated multi-domain, task-oriented conversations between a "
            "human and a virtual assistant."
        ),
    },
]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import time
import uuid
import os
from unittest import mock
from parlai.mturk.core.dev.socket_manager import Packet, SocketManager
from parlai.mturk.core.dev.worlds import MTurkOnboardWorld, MTurkTaskWorld
from parlai.mturk.core.dev.agents import AssignState
from parlai.mturk.core.dev.mturk_manager import MTurkManager
from parlai.core.params import ParlaiParser

import parlai.mturk.core.dev.mturk_manager as MTurkManagerFile
import parlai.mturk.core.dev.data_model as data_model
import parlai.mturk.core.dev.shared_utils as shared_utils
import threading
from websocket_server import WebsocketServer
import json

parent_dir = os.path.dirname(os.path.abspath(__file__))
MTurkManagerFile.parent_dir = os.path.dirname(os.path.abspath(__file__))

# Lets ignore the logging part
MTurkManagerFile.shared_utils.print_and_log = mock.MagicMock()

TEST_WORKER_ID_1 = 'TEST_WORKER_ID_1'
TEST_WORKER_ID_2 = 'TEST_WORKER_ID_2'
TEST_ASSIGNMENT_ID_1 = 'TEST_ASSIGNMENT_ID_1'
TEST_ASSIGNMENT_ID_2 = 'TEST_ASSIGNMENT_ID_2'
TEST_ASSIGNMENT_ID_3 = 'TEST_ASSIGNMENT_ID_3'
TEST_HIT_ID_1 = 'TEST_HIT_ID_1'
TEST_HIT_ID_2 = 'TEST_HIT_ID_2'
TEST_CONV_ID_1 = 'TEST_CONV_ID_1'
FAKE_ID = 'BOGUS'

MESSAGE_ID_1 = 'MESSAGE_ID_1'
MESSAGE_ID_2 = 'MESSAGE_ID_2'
MESSAGE_ID_3 = 'MESSAGE_ID_3'
MESSAGE_ID_4 = 'MESSAGE_ID_4'
COMMAND_ID_1 = 'COMMAND_ID_1'

MESSAGE_TYPE = data_model.MESSAGE_TYPE_ACT
COMMAND_TYPE = data_model.MESSAGE_TYPE_COMMAND

MESSAGE_1 = {'message_id': MESSAGE_ID_1, 'type': MESSAGE_TYPE}
MESSAGE_2 = {'message_id': MESSAGE_ID_2, 'type': MESSAGE_TYPE}
COMMAND_1 = {'message_id': COMMAND_ID_1, 'type': COMMAND_TYPE}

AGENT_ID = 'AGENT_ID'

ACT_1 = {'text': 'THIS IS A MESSAGE', 'id': AGENT_ID}
ACT_2 = {'text': 'THIS IS A MESSAGE AGAIN', 'id': AGENT_ID}

active_statuses = [
    AssignState.STATUS_NONE,
    AssignState.STATUS_ONBOARDING,
    AssignState.STATUS_WAITING,
    AssignState.STATUS_IN_TASK,
]
complete_statuses = [
    AssignState.STATUS_DONE,
    AssignState.STATUS_DISCONNECT,
    AssignState.STATUS_PARTNER_DISCONNECT,
    AssignState.STATUS_PARTNER_DISCONNECT_EARLY,
    AssignState.STATUS_EXPIRED,
    AssignState.STATUS_RETURNED,
]
statuses = active_statuses + complete_statuses

TASK_GROUP_ID_1 = 'TASK_GROUP_ID_1'

SocketManager.DEF_MISSED_PONGS = 1
SocketManager.DEF_DEAD_TIME = 0.4

shared_utils.THREAD_SHORT_SLEEP = 0.05
shared_utils.THREAD_MEDIUM_SLEEP = 0.15

MTurkManagerFile.WORLD_START_TIMEOUT = 2


TOPIC_ARN = 'topic_arn'
QUALIFICATION_ID = 'qualification_id'
HIT_TYPE_ID = 'hit_type_id'
MTURK_PAGE_URL = 'mturk_page_url'
FAKE_HIT_ID = 'fake_hit_id'


class TestMTurkWorld(MTurkTaskWorld):
    def __init__(self, workers, use_episode_done):
        self.workers = workers

        def episode_done():
            return use_episode_done()

        self.episode_done = episode_done

    def parley(self):
        for worker in self.workers:
            worker.assert_connected()
        time.sleep(0.5)

    def shutdown(self):
        for worker in self.workers:
            worker.shutdown()


class TestMTurkOnboardWorld(MTurkOnboardWorld):
    def __init__(self, mturk_agent, use_episode_done):
        self.mturk_agent = mturk_agent

        def episode_done():
            return use_episode_done()

        self.episode_done = episode_done

    def parley(self):
        self.mturk_agent.assert_connected()
        time.sleep(0.5)


def assert_equal_by(val_func, val, max_time):
    start_time = time.time()
    while val_func() != val:
        assert (
            time.time() - start_time < max_time
        ), "Value was not attained in specified time, last {}".format(val_func())
        time.sleep(0.1)


class MockSocket:
    def __init__(self):
        self.last_messages = {}
        self.connected = False
        self.disconnected = False
        self.closed = False
        self.ws = None
        self.fake_workers = []
        self.port = None
        self.launch_socket()
        self.handlers = {}
        while self.ws is None:
            time.sleep(0.05)
        time.sleep(1)

    def send(self, packet):
        self.ws.send_message_to_all(packet)

    def close(self):
        if not self.closed:
            self.ws.server_close()
            self.ws.shutdown()
            self.closed = True

    def do_nothing(self, *args):
        pass

    def launch_socket(self):
        def on_message(client, server, message):
            if self.closed:
                raise Exception('Socket is already closed...')
            if message == '':
                return
            packet_dict = json.loads(message)
            if packet_dict['content']['id'] == 'WORLD_ALIVE':
                self.ws.send_message(client, json.dumps({'type': 'conn_success'}))
                self.connected = True
            elif packet_dict['type'] == data_model.WORLD_PING:
                pong = packet_dict['content'].copy()
                pong['type'] = 'pong'
                self.ws.send_message(
                    client,
                    json.dumps({'type': data_model.SERVER_PONG, 'content': pong}),
                )
            if 'receiver_id' in packet_dict['content']:
                receiver_id = packet_dict['content']['receiver_id']
                use_func = self.handlers.get(receiver_id, self.do_nothing)
                use_func(packet_dict['content'])

        def on_connect(client, server):
            pass

        def on_disconnect(client, server):
            self.disconnected = True

        def run_socket(*args):
            port = 3030
            while self.port is None:
                try:
                    self.ws = WebsocketServer(port, host='127.0.0.1')
                    self.port = port
                except OSError:
                    port += 1
            self.ws.set_fn_client_left(on_disconnect)
            self.ws.set_fn_new_client(on_connect)
            self.ws.set_fn_message_received(on_message)
            self.ws.run_forever()

        self.listen_thread = threading.Thread(
            target=run_socket, name='Fake-Socket-Thread'
        )
        self.listen_thread.daemon = True
        self.listen_thread.start()


class MockAgent(object):
    """
    Class that pretends to be an MTurk agent interacting through the webpage by
    simulating the same commands that are sent from the core.html file.

    Exposes methods to use for testing and checking status
    """

    def __init__(self, hit_id, assignment_id, worker_id, task_group_id):
        self.conversation_id = None
        self.id = None
        self.assignment_id = assignment_id
        self.hit_id = hit_id
        self.worker_id = worker_id
        self.some_agent_disconnected = False
        self.disconnected = False
        self.task_group_id = task_group_id
        self.ws = None
        self.ready = False
        self.wants_to_send = False
        self.message_packet = []

    def send_packet(self, packet):
        def callback(*args):
            pass

        event_name = data_model.MESSAGE_BATCH
        self.ws.send(json.dumps({'type': event_name, 'content': packet.as_dict()}))

    def register_to_socket(self, ws, on_msg=None):
        if on_msg is None:

            def on_msg(packet):
                self.message_packet.append(packet)
                if packet.type == data_model.AGENT_STATE_CHANGE:
                    if 'conversation_id' in packet.data:
                        self.conversation_id = packet.data['conversation_id']
                    if 'agent_id' in packet.data:
                        self.id = packet.data['agent_id']

        handler = self.make_packet_handler(on_msg)
        self.ws = ws
        self.ws.handlers[self.worker_id] = handler

    def make_packet_handler(self, on_msg):
        """
        A packet handler.
        """

        def handler_mock(pkt):
            if pkt['type'] == data_model.WORLD_MESSAGE:
                packet = Packet.from_dict(pkt)
                on_msg(packet)
            elif pkt['type'] == data_model.MESSAGE_BATCH:
                packet = Packet.from_dict(pkt)
                on_msg(packet)
            elif pkt['type'] == data_model.AGENT_STATE_CHANGE:
                packet = Packet.from_dict(pkt)
                on_msg(packet)
            elif pkt['type'] == data_model.AGENT_ALIVE:
                raise Exception('Invalid alive packet {}'.format(pkt))
            else:
                raise Exception(
                    'Invalid Packet type {} received in {}'.format(pkt['type'], pkt)
                )

        return handler_mock

    def build_and_send_packet(self, packet_type, data):
        msg_id = str(uuid.uuid4())
        msg = {
            'id': msg_id,
            'type': packet_type,
            'sender_id': self.worker_id,
            'assignment_id': self.assignment_id,
            'conversation_id': self.conversation_id,
            'receiver_id': '[World_' + self.task_group_id + ']',
            'data': data,
        }

        if packet_type == data_model.MESSAGE_BATCH:
            msg['data'] = {
                'messages': [
                    {
                        'id': msg_id,
                        'type': packet_type,
                        'sender_id': self.worker_id,
                        'assignment_id': self.assignment_id,
                        'conversation_id': self.conversation_id,
                        'receiver_id': '[World_' + self.task_group_id + ']',
                        'data': data,
                    }
                ]
            }
        self.ws.send(json.dumps({'type': packet_type, 'content': msg}))
        return msg['id']

    def send_message(self, text):
        data = {
            'text': text,
            'id': self.id,
            'message_id': str(uuid.uuid4()),
            'episode_done': False,
        }

        self.wants_to_send = False
        return self.build_and_send_packet(data_model.MESSAGE_BATCH, data)

    def send_disconnect(self):
        data = {
            'hit_id': self.hit_id,
            'assignment_id': self.assignment_id,
            'worker_id': self.worker_id,
            'conversation_id': self.conversation_id,
            'connection_id': '{}_{}'.format(self.worker_id, self.assignment_id),
        }

        return self.build_and_send_packet(data_model.AGENT_DISCONNECT, data)

    def send_alive(self):
        data = {
            'hit_id': self.hit_id,
            'assignment_id': self.assignment_id,
            'worker_id': self.worker_id,
            'conversation_id': self.conversation_id,
        }
        return self.build_and_send_packet(data_model.AGENT_ALIVE, data)


class TestMTurkManagerWorkflows(unittest.TestCase):
    """
    Various test cases to replicate a whole mturk workflow.
    """

    def setUp(self):
        patcher = mock.patch('builtins.input', return_value='y')
        self.addCleanup(patcher.stop)
        patcher.start()
        # Mock functions that hit external APIs and such
        self.server_utils = MTurkManagerFile.server_utils
        self.mturk_utils = MTurkManagerFile.mturk_utils
        self.server_utils.setup_server = mock.MagicMock(
            return_value='https://127.0.0.1'
        )
        self.server_utils.setup_legacy_server = mock.MagicMock(
            return_value='https://127.0.0.1'
        )
        self.server_utils.delete_server = mock.MagicMock()
        self.mturk_utils.setup_aws_credentials = mock.MagicMock()
        self.mturk_utils.calculate_mturk_cost = mock.MagicMock(return_value=1)
        self.mturk_utils.check_mturk_balance = mock.MagicMock(return_value=True)
        self.mturk_utils.create_hit_config = mock.MagicMock()
        self.mturk_utils.setup_sns_topic = mock.MagicMock(return_value=TOPIC_ARN)
        self.mturk_utils.delete_sns_topic = mock.MagicMock()
        self.mturk_utils.delete_qualification = mock.MagicMock()
        self.mturk_utils.find_or_create_qualification = mock.MagicMock(
            return_value=QUALIFICATION_ID
        )
        self.mturk_utils.find_qualification = mock.MagicMock(
            return_value=QUALIFICATION_ID
        )
        self.mturk_utils.give_worker_qualification = mock.MagicMock()
        self.mturk_utils.remove_worker_qualification = mock.MagicMock()
        self.mturk_utils.create_hit_type = mock.MagicMock(return_value=HIT_TYPE_ID)
        self.mturk_utils.subscribe_to_hits = mock.MagicMock()
        self.mturk_utils.create_hit_with_hit_type = mock.MagicMock(
            return_value=(MTURK_PAGE_URL, FAKE_HIT_ID, 'MTURK_HIT_DATA')
        )
        self.mturk_utils.get_mturk_client = mock.MagicMock(
            return_value=mock.MagicMock()
        )

        self.onboarding_agents = {}
        self.worlds_agents = {}

        # Set up an MTurk Manager and get it ready for accepting workers
        self.fake_socket = MockSocket()
        time.sleep(0.1)
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args()
        self.opt['task'] = 'unittest'
        self.opt['frontend_version'] = 1
        self.opt['assignment_duration_in_seconds'] = 1
        self.opt['hit_title'] = 'test_hit_title'
        self.opt['hit_description'] = 'test_hit_description'
        self.opt['task_description'] = 'test_task_description'
        self.opt['hit_keywords'] = 'test_hit_keywords'
        self.opt['reward'] = 0.1
        self.opt['is_debug'] = True
        self.opt['log_level'] = 0
        self.opt['num_conversations'] = 1
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        self.mturk_manager.port = self.fake_socket.port
        self.mturk_manager.setup_server()
        self.mturk_manager.start_new_run()
        self.mturk_manager.ready_to_accept_workers()
        self.mturk_manager.set_get_onboard_world(self.get_onboard_world)
        self.mturk_manager.create_hits()

        def assign_worker_roles(workers):
            workers[0].id = 'mturk_agent_1'
            workers[1].id = 'mturk_agent_2'

        def run_task_wait():
            self.mturk_manager.start_task(
                lambda w: True, assign_worker_roles, self.get_task_world
            )

        self.task_thread = threading.Thread(target=run_task_wait)
        self.task_thread.start()

        self.agent_1 = MockAgent(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_1, TEST_WORKER_ID_1, TASK_GROUP_ID_1
        )
        self.agent_1_2 = MockAgent(
            TEST_HIT_ID_1, TEST_ASSIGNMENT_ID_3, TEST_WORKER_ID_1, TASK_GROUP_ID_1
        )
        self.agent_2 = MockAgent(
            TEST_HIT_ID_2, TEST_ASSIGNMENT_ID_2, TEST_WORKER_ID_2, TASK_GROUP_ID_1
        )

    def tearDown(self):
        for key in self.worlds_agents.keys():
            self.worlds_agents[key] = True
        self.mturk_manager.shutdown()
        self.fake_socket.close()
        if self.task_thread.isAlive():
            self.task_thread.join()

    def get_onboard_world(self, mturk_agent):
        self.onboarding_agents[mturk_agent.worker_id] = False

        def episode_done():
            return not (
                (mturk_agent.worker_id in self.onboarding_agents)
                and (self.onboarding_agents[mturk_agent.worker_id] is False)
            )

        return TestMTurkOnboardWorld(mturk_agent, episode_done)

    def get_task_world(self, mturk_manager, opt, workers):
        for worker in workers:
            self.worlds_agents[worker.worker_id] = False

        def episode_done():
            for worker in workers:
                if self.worlds_agents[worker.worker_id] is False:
                    return False
            return True

        return TestMTurkWorld(workers, episode_done)

    def alive_agent(self, agent):
        agent.register_to_socket(self.fake_socket)
        agent.send_alive()
        time.sleep(0.3)

    def test_successful_convo(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        agent_1_object.set_completed_act({})
        agent_2_object.set_completed_act({})
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        # Assert conversation is complete for manager and agents
        assert_equal_by(lambda: manager.completed_conversations, 1, 2)

    def test_disconnect_end(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Disconnect agent
        agent_2.send_disconnect()
        assert_equal_by(
            agent_1_object.get_status, AssignState.STATUS_PARTNER_DISCONNECT, 3
        )
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DISCONNECT, 3)
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True

        # assert conversation not marked as complete
        self.assertEqual(manager.completed_conversations, 0)
        agent_1_object.set_completed_act({})

    def test_expire_onboarding(self):
        manager = self.mturk_manager

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 10)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)

        manager._expire_onboarding_pool()
        self.onboarding_agents[agent_1.worker_id] = True

        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_EXPIRED)

    def test_attempt_break_unique(self):
        manager = self.mturk_manager
        unique_worker_qual = 'is_unique_qual'
        manager.is_unique = True
        manager.opt['unique_qual_name'] = unique_worker_qual
        manager.unique_qual_name = unique_worker_qual

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        agent_1_object.set_completed_act({})
        agent_2_object.set_completed_act({})
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        # Assert conversation is complete for manager and agents
        assert_equal_by(lambda: manager.completed_conversations, 1, 2)

        # ensure qualification was 'granted'
        self.mturk_utils.find_qualification.assert_called_with(
            unique_worker_qual, manager.is_sandbox
        )
        self.mturk_utils.give_worker_qualification.assert_any_call(
            agent_1.worker_id, QUALIFICATION_ID, None, manager.is_sandbox
        )
        self.mturk_utils.give_worker_qualification.assert_any_call(
            agent_2.worker_id, QUALIFICATION_ID, None, manager.is_sandbox
        )

        # Try to alive with the first agent a second time
        agent_1_2 = self.agent_1_2
        self.alive_agent(agent_1_2)
        assert_equal_by(lambda: agent_1_2.worker_id in self.onboarding_agents, True, 2)
        agent_1_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_1_2.assignment_id
        )

        # No worker should be created for a unique task
        self.assertIsNone(agent_1_2_object)

    def test_break_multi_convo(self):
        manager = self.mturk_manager
        manager.opt['allowed_conversations'] = 1

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        assert_equal_by(lambda: agent_1.worker_id in self.onboarding_agents, True, 2)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_1.worker_id])
        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_1.worker_id] = True
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        # Alive second agent
        agent_2 = self.agent_2
        self.alive_agent(agent_2)
        assert_equal_by(lambda: agent_2.worker_id in self.onboarding_agents, True, 2)
        agent_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_2.assignment_id
        )
        self.assertFalse(self.onboarding_agents[agent_2.worker_id])
        self.assertEqual(agent_2_object.get_status(), AssignState.STATUS_ONBOARDING)
        self.onboarding_agents[agent_2.worker_id] = True

        # Assert agents move to task
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_IN_TASK, 2)
        assert_equal_by(lambda: agent_2.worker_id in self.worlds_agents, True, 2)
        self.assertIn(agent_1.worker_id, self.worlds_agents)

        # Attempt to start a new conversation with duplicate worker 1
        agent_1_2 = self.agent_1_2
        self.alive_agent(agent_1_2)
        assert_equal_by(lambda: agent_1_2.worker_id in self.onboarding_agents, True, 2)
        agent_1_2_object = manager.worker_manager.get_agent_for_assignment(
            agent_1_2.assignment_id
        )

        # No worker should be created for a unique task
        self.assertIsNone(agent_1_2_object)

        # Complete agents
        self.worlds_agents[agent_1.worker_id] = True
        self.worlds_agents[agent_2.worker_id] = True
        agent_1_object.set_completed_act({})
        agent_2_object.set_completed_act({})
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_DONE, 2)
        assert_equal_by(agent_2_object.get_status, AssignState.STATUS_DONE, 2)

        assert_equal_by(lambda: manager.completed_conversations, 1, 2)

    def test_no_onboard_expire_waiting(self):
        manager = self.mturk_manager
        manager.set_get_onboard_world(None)

        # Alive first agent
        agent_1 = self.agent_1
        self.alive_agent(agent_1)
        agent_1_object = manager.worker_manager.get_agent_for_assignment(
            agent_1.assignment_id
        )
        assert_equal_by(agent_1_object.get_status, AssignState.STATUS_WAITING, 2)

        manager._expire_agent_pool()

        self.assertEqual(agent_1_object.get_status(), AssignState.STATUS_EXPIRED)


if __name__ == '__main__':
    unittest.main(buffer=True)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import DialogTeacher, FixedDialogTeacher
from .build import build

import csv
import random
import os


class UbuntuTeacher(DialogTeacher):
    """
    This teacher inherits from the core Dialog Teacher, which just requires it to define
    an iterator over its data `setup_data` in order to inherit basic metrics, a default
    `act` function, and enables Hogwild training with shared memory with no extra work.
    """

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)
        opt['datafile'] = os.path.join(
            opt['datapath'], 'Ubuntu', opt['datatype'].split(':')[0] + '.csv'
        )
        super().__init__(opt, shared)

    def setup_data(self, path):
        print('loading: ' + path)
        with open(path, 'r', newline='') as read:
            csv_read = csv.reader(read)
            next(csv_read)  # eat header

            for line in csv_read:
                fields = [
                    s.replace('__eou__', '.').replace('__eot__', '\n').strip()
                    for s in line
                ]
                context = fields[0]
                response = fields[1]
                cands = None
                if len(fields) > 3:
                    cands = [fields[i] for i in range(2, len(fields))]
                    cands.append(response)
                    random.shuffle(cands)
                yield (context, [response], None, cands), True


class MultiturnTeacher(FixedDialogTeacher):
    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        opt['datafile'] = os.path.join(
            opt['datapath'], 'Ubuntu', opt['datatype'].split(':')[0] + '.csv'
        )
        super().__init__(opt, shared)

        if shared:
            self.data = shared['data']
        else:
            build(opt)
            fold = opt.get('datatype', 'train').split(':')[0]
            self._setup_data(fold)

        self.num_exs = sum(len(d) for d in self.data)

        # we learn from both sides of every conversation
        self.num_eps = 2 * len(self.data)
        self.reset()

    def num_episodes(self):
        return self.num_eps

    def num_examples(self):
        return self.num_exs

    def _setup_data(self, fold):
        self.data = []
        fpath = os.path.join(self.opt['datapath'], 'Ubuntu', fold + '.csv')
        print('loading: ' + fpath)
        with open(fpath, 'r', newline='') as read:
            csv_read = csv.reader(read)
            next(csv_read)  # eat header

            for line in csv_read:
                fields = line[0].strip().split('__eou__ __eot__')
                fields.append(line[1].strip())
                dialog = []
                for field in fields:
                    if field != '':
                        dialog.append(field.replace('__eou__', '.').strip())
                if len(dialog) > 2:
                    self.data.append(dialog)

    def get(self, episode_idx, entry_idx=0):
        # Sometimes we're speaker 1 and sometimes we're speaker 2
        speaker_id = episode_idx % 2
        full_eps = self.data[episode_idx // 2]

        entries = full_eps
        their_turn = entries[speaker_id + 2 * entry_idx]
        my_turn = entries[1 + speaker_id + 2 * entry_idx]

        episode_done = 1 + speaker_id + 2 * entry_idx >= len(full_eps) - 2

        action = {'text': their_turn, 'labels': [my_turn], 'episode_done': episode_done}
        return action

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared


class DefaultTeacher(UbuntuTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

task_config = {}

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config[
    'hit_title'
] = 'Navigating New York City v2 -- 40c bonus upon successful completion.'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config[
    'hit_description'
] = 'You\'ll be playing the role of a Tourist or a Guide, navigating New York City. Bonus of 40c upon successful completion. You can keep accepting new HITs.'


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat, navigation, streetview'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config['task_description'] = ''



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FixedDialogTeacher
from .build import build
import os
import json


class ELI5Teacher(FixedDialogTeacher):
    """
    ELI5 Teacher, taken from https://github.com/facebookresearch/ELI5.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        build(opt)
        self.id = 'eli5'
        self.messages = self.load_eli5(opt)
        self.reset()

    @staticmethod
    def add_cmdline_args(argparser):
        group = argparser.add_argument_group('ELI5 Knowledge arguments')
        group.add_argument(
            '--knowledge',
            type='bool',
            default=True,
            help='Whether to include supporting document knowledge',
        )

    def load_eli5(self, opt):
        """
        Load data based on data split.
        """
        dp = opt['datapath']
        dt = opt['datatype'].split(':')[0]
        eli_path = "eli5/processed_data/selected_15_1/explainlikeimfive_"
        fname = os.path.join(dp, eli_path + dt + ".json")
        if not os.path.isfile(fname):
            raise FileNotFoundError(
                f"{fname} not found. Please follow the instructions found at "
                "https://github.com/facebookresearch/ParlAI/tree/master/parlai/tasks/eli5/README.md"
                " to construct the dataset."
            )
        opt['datafile'] = fname
        with open(fname) as json_file:
            data = json.load(json_file)
        ds = []
        for d in data:
            if self.opt['knowledge']:
                text = d['document'] + "\n" + d['question']
            else:
                text = d['question']
            act = {
                'id': 'eli5',
                'text': text,
                'labels': [d['answer']],
                'episode_done': True,
            }
            ds.append(act)
        return ds

    def get(self, episode_idx, entry_idx=0):
        return self.messages[episode_idx]

    def num_examples(self):
        return len(self.messages)

    def num_episodes(self):
        return len(self.messages)


class DefaultTeacher(ELI5Teacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, Any

task_config: Dict[str, Any] = {}


task_config['frontend_version'] = 1

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Ask, answer, and evaluate numeric questions'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config[
    'hit_description'
] = 'Perform one of three roles to get accurate numeric questions.'


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,question,answer'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = '''
In this task, you'll need to ask, answer, or evaluate numeric questions.<br><br>
<b>As the asker:</b> please only ask questions that can be answered with a
number. Prefer to use natural language questions rather than typed
up equations.<br>
<b>As the answerer:</b> feel free to use external resources to come up with
answers to questions if you don't know the answer.<br>
<b>As the evaluator:</b> please evaluate based on accuracy of the answers as
well as how answerable the questions are.

If you are ready, please click "Accept HIT" to start this task.
'''
